import os
from argparse import ArgumentParser
from functools import partial

from elasticai.creator.vhdl.lstm_testbench_generator import LSTMCellTestBench
from elasticai.creator.vhdl.number_representations import (
    float_values_to_fixed_point,
    int_values_to_fixed_point,
)
from elasticai.creator.vhdl.vhdl_formatter.vhdl_formatter import format_vhdl


def main() -> None:
    arg_parser = ArgumentParser()
    arg_parser.add_argument(
        "--path",
        help="path to folder for generated vhd files",
        required=True,
    )
    args = arg_parser.parse_args()
    if not os.path.isdir(args.path):
        os.mkdir(args.path)
    component_name = "lstm_cell_tb"
    file_path = os.path.join(args.path, f"{component_name}.vhd")

    fp_args = dict(total_bits=16, frac_bits=8)
    ints_to_fp = partial(int_values_to_fixed_point, **fp_args)
    floats_to_fp = partial(float_values_to_fixed_point, **fp_args)

    lstm_cell = LSTMCellTestBench(
        input_size=5,
        hidden_size=20,
        test_x_h_data=ints_to_fp(
            [
                0x018A,
                0xFFB5,
                0xFDD3,
                0x0091,
                0xFEEB,
                0x0099,
                0xFE72,
                0xFFA9,
                0x01DA,
                0xFFC9,
                0xFF42,
                0x0090,
                0x0042,
                0xFFD4,
                0xFF53,
                0x00F0,
                0x007D,
                0x0134,
                0x0015,
                0xFECD,
                0xFFFF,
                0xFF7C,
                0xFFB2,
                0xFE6C,
                0x01B4,
                0x0000,
                0x0000,
                0x0000,
                0x0000,
                0x0000,
                0x0000,
                0x0000,
            ]
        ),
        test_c_data=ints_to_fp(
            [
                0x0034,
                0xFF8D,
                0xFF6E,
                0xFF72,
                0xFEE0,
                0xFFAF,
                0xFEE9,
                0xFFEB,
                0xFFE9,
                0x00AF,
                0xFF2A,
                0x0000,
                0xFF40,
                0x002F,
                0x009F,
                0x00A3,
                0xFFC2,
                0x024D,
                0xFE1F,
                0xFFF4,
                0x0000,
                0x0000,
                0x0000,
                0x0000,
                0x0000,
                0x0000,
                0x0000,
                0x0000,
                0x0000,
                0x0000,
                0x0000,
                0x0000,
            ]
        ),
        h_out=floats_to_fp(
            [
                34,
                -80,
                -32,
                -28,
                -88,
                11,
                -60,
                6,
                -16,
                18,
                -32,
                46,
                -77,
                15,
                70,
                27,
                13,
                112,
                -126,
                3,
            ]
        ),
        component_name=component_name,
    )
    lstm_cell_code = lstm_cell()

    with open(file_path, "w") as writer:
        for line in lstm_cell_code:
            writer.write(line + "\n")

    format_vhdl(file_path=file_path)


if __name__ == "__main__":
    main()



from functools import partial
from io import StringIO

from torch.nn import Conv1d, MaxPool1d, Sequential

from elasticai.creator.precomputation.input_domains import (
    create_codomain_for_1d_conv,
    create_codomain_for_depthwise_1d_conv,
)
from elasticai.creator.precomputation.precomputation import (
    get_precomputations_from_direct_children,
    precomputable,
)
from elasticai.creator.qat.layers import Binarize, QConv1d

model = Sequential(
    QConv1d(
        in_channels=2,
        out_channels=2,
        kernel_size=(3,),
        quantizer=Binarize(),
    ),
    MaxPool1d(kernel_size=2),
    Binarize(),
    precomputable(
        Sequential(
            Conv1d(
                in_channels=2,
                out_channels=2,
                kernel_size=(4,),
                groups=2,
            ),
            Binarize(),
        ),
        input_generator=partial(
            create_codomain_for_depthwise_1d_conv, shape=(2, 4), items=[-1, 1]
        ),
        input_shape=(2, 2),
    ),
    MaxPool1d(kernel_size=2),
    precomputable(
        Sequential(
            Conv1d(
                in_channels=2,
                out_channels=3,
                kernel_size=(1,),
            ),
            Binarize(),
        ),
        input_generator=partial(
            create_codomain_for_1d_conv, shape=(1, 2), items=[-1, 1]
        ),
        input_shape=(6, 1),
    ),
)

model.eval()

with StringIO() as f:
    for precompute in get_precomputations_from_direct_children(model):
        precompute()
        precompute.dump_to(f)
    print(f.getvalue())



import unittest

import torch
from torch.nn.utils import parametrize

from elasticai.creator.precomputation.breakdown import (
    BreakdownConv2dBlock,
    depthwisePointwiseBreakdownConv2dBlock,
    generate_conv2d_sequence_with_width,
)
from elasticai.creator.qat.blocks import Conv2dBlock
from elasticai.creator.qat.layers import Binarize, ChannelShuffle
from elasticai.creator.qat.masks import FixedOffsetMask4d


class BreakdownTest(unittest.TestCase):
    def compare_models_and_weight_shape(self, expected, actual):
        self.assertTrue(len(expected) == len(actual), "number of layers")
        for layer_expected, layer_actual in zip(expected, actual):
            if not hasattr(
                layer_expected, "parametrizations"
            ):  # parametrizations seems to create a dynamic class, impossible to compare instances
                self.assertTrue(
                    isinstance(layer_expected, type(layer_expected)),
                    f"layer types differ expected:{layer_expected} actual:{layer_actual}",
                )
            if hasattr(layer_actual, "conv2d"):
                self.assertSequenceEqual(
                    layer_actual.conv2d.weight.shape,
                    layer_expected.conv2d.weight.shape,
                    f"layer shapes differ expected:{layer_expected.conv2d.weight.shape} actual:{layer_actual.conv2d.weight.shape}",
                )
            if hasattr(layer_actual, "groups"):
                self.assertEqual(
                    layer_actual.groups,
                    layer_expected.groups,
                    f"groups differ expected:{layer_expected.groups} actual:{layer_actual.groups}",
                )

    def test_generate_conv2d_sequence_with_width_base(self):
        layers = generate_conv2d_sequence_with_width(
            in_channels=2,
            out_channels=4,
            activation=torch.nn.Identity(),
            weight_quantization=Binarize(),
            kernel_size=3,
            channel_width=2,
        )
        expected = torch.nn.Sequential(
            Conv2dBlock(
                in_channels=2,
                out_channels=4,
                activation=torch.nn.Identity(),
                conv_quantizer=Binarize(),
                kernel_size=3,
            ),
            ChannelShuffle(groups=1),
        )
        self.compare_models_and_weight_shape(layers, expected)

    def test_generate_conv2d_sequence_with_width_more_complex(self):
        layers = generate_conv2d_sequence_with_width(
            in_channels=4,
            out_channels=2,
            activation=torch.nn.Identity(),
            weight_quantization=Binarize(),
            kernel_size=3,
            channel_width=2,
        )
        expected = torch.nn.Sequential(
            Conv2dBlock(
                in_channels=4,
                out_channels=4,
                activation=torch.nn.Identity(),
                conv_quantizer=Binarize(),
                kernel_size=3,
                groups=2,
            ),
            ChannelShuffle(groups=2),
            Conv2dBlock(
                in_channels=4,
                out_channels=2,
                activation=torch.nn.Identity(),
                conv_quantizer=Binarize(),
                kernel_size=3,
                groups=2,
            ),
        )
        self.compare_models_and_weight_shape(layers, expected)

    def test_generate_conv2d_sequence_with_width_more_than_2_last_groups_unequal_out_channels(
        self,
    ):
        layers = generate_conv2d_sequence_with_width(
            in_channels=256,
            out_channels=256,
            activation=torch.nn.Identity(),
            weight_quantization=Binarize(),
            kernel_size=3,
            channel_width=8,
        )
        expected = torch.nn.Sequential(
            Conv2dBlock(
                in_channels=256,
                out_channels=256 * 32,
                activation=torch.nn.Identity(),
                conv_quantizer=Binarize(),
                kernel_size=3,
                groups=32,
            ),
            ChannelShuffle(groups=32),
            Conv2dBlock(
                in_channels=256 * 32,
                out_channels=256 * 4,
                activation=torch.nn.Identity(),
                conv_quantizer=Binarize(),
                kernel_size=3,
                groups=256 * 4,
            ),
            Conv2dBlock(
                in_channels=256 * 4,
                out_channels=256,
                activation=torch.nn.Identity(),
                conv_quantizer=Binarize(),
                kernel_size=3,
                groups=128,
            ),
        )
        self.compare_models_and_weight_shape(layers, expected)

    def test_generate_conv2d_sequence_with_width_more_than_2_last_groups_equal_out_channels(
        self,
    ):
        layers = generate_conv2d_sequence_with_width(
            in_channels=64,
            out_channels=256,
            activation=torch.nn.Identity(),
            weight_quantization=Binarize(),
            kernel_size=3,
            channel_width=8,
        )
        expected = torch.nn.Sequential(
            Conv2dBlock(
                in_channels=64,
                out_channels=256 * 8,
                activation=torch.nn.Identity(),
                conv_quantizer=Binarize(),
                kernel_size=3,
                groups=8,
            ),
            ChannelShuffle(groups=8),
            Conv2dBlock(
                in_channels=256 * 8,
                out_channels=256,
                activation=torch.nn.Identity(),
                conv_quantizer=Binarize(),
                kernel_size=3,
                groups=256,
            ),
        )
        self.compare_models_and_weight_shape(layers, expected)

    def test_pointwise_breakdown_forward(self):
        layer = depthwisePointwiseBreakdownConv2dBlock(
            in_channels=4,
            out_channels=8,
            pointwise_channel_width=2,
            kernel_size=3,
            activation=Binarize(),
            pointwise_activation=Binarize(),
        )
        test_input = torch.ones((2, 4, 3, 3))
        output = layer(test_input)
        self.assertSequenceEqual(output.shape, [2, 8, 1, 1])

    def test_BreakdownConv2d_block(self):
        layers = BreakdownConv2dBlock(
            in_channels=32,
            out_channels=16,
            kernel_size=2,
            groups=4,
            activation=torch.nn.Identity(),
            pointwise_activation=torch.nn.Identity(),
            conv_quantizer=None,
            pointwise_quantizer=None,
            stride=1,
            padding=0,
            dilation=1,
            bias=True,
            padding_mode="zeros",
        )
        expected = torch.nn.Sequential(
            Conv2dBlock(
                in_channels=32,
                out_channels=16 * 4 * 2,
                activation=torch.nn.Identity(),
                conv_quantizer=None,
                kernel_size=2,
                groups=4,
            ),
            ChannelShuffle(groups=4),
            Conv2dBlock(
                in_channels=16 * 4 * 2,
                out_channels=16,
                activation=torch.nn.Identity(),
                conv_quantizer=None,
                kernel_size=1,
                groups=16,
            ),
        )
        mask = FixedOffsetMask4d(
            out_channels=16 * 4 * 2,
            in_channels=32,
            kernel_size=2,
            offset_axis=2,
            axis_width=1,
            groups=4,
        )
        parametrize.register_parametrization(expected[0].conv2d, "weight", mask)
        self.compare_models_and_weight_shape(
            list(layers.modules())[1:], list(expected.modules())[1:]
        )


if __name__ == "__main__":
    unittest.main()



import itertools
from typing import Iterable

import numpy as np
import torch

from elasticai.creator.mlframework import Module, Tensor


def create_input_data(
    input_dim: list[int], domain: list[int | float], dtype="float16"
) -> np.ndarray:
    """The function creates possible inputs for a submodel given input dimensions .
    Args:
        input_dim (list[int]): The dimensions of the input, can be a tuple too.
        domain (list[int | float]): The possible values of the input

    Returns:
        numpy.ndarray: An array of possible inputs
    """
    input_dim = list(input_dim)
    input_dim = [x for x in input_dim if x is not None]
    domain = np.unique(domain)
    n_values = int(len(domain) ** np.prod(input_dim))
    table = np.zeros([n_values] + input_dim, dtype=dtype)

    for vector, i in zip(
        itertools.product(domain, repeat=int(np.prod(input_dim))), range(n_values)
    ):
        table[i] = np.reshape(np.asarray(vector), input_dim)

    return table


def get_cartesian_product_from_items(
    length: int, items: Iterable, dtype="float16"
) -> np.ndarray:
    return np.array(tuple(itertools.product(items, repeat=length)), dtype=dtype)


def construct_codomain_from_elements(
    shape: tuple[int, ...], codomain_elements: Iterable, dtype="float16"
) -> np.ndarray:
    """Build the numpy array containing all combinations that can be built from `items` resulting in the desired `shape`.

    `items` can either be a flat iterable of scalar values or already structure that is convertible to a nested numpy array.
    In the latter case the last dimensions of the requested `shape` are expected to match the shape of `items` exactly.
    Otherwise a `ValueError` will be raised.
    E.g.:
    ```
      construct_domain_from_items((3, 2), ((1, 1), (0, 0))) # is fine
      construct_domain_from_items((2, 3), ((1, 1), (0, 0))) # will raise a ValueError
    ```
    """
    result = np.array(codomain_elements)

    shape_we_need_to_build = shape[
        : _calculate_the_rank_index_that_shape_sizes_match_up_to(
            requested_shape=shape, provided_shape=result.shape[1:]
        )
    ]
    for size in shape_we_need_to_build:
        result = get_cartesian_product_from_items(
            length=size, items=result, dtype=dtype
        )
    return result


def create_io_table(
    inputs: np.ndarray, outputs: list[Tensor], channel_wise=True
) -> dict | list[dict]:
    """The function creates io tables given an array of inputs and a model .
    Args:
        inputs (np.ndarray): The inputs
        outputs (list[Tensor]): The outputs
        channel_wise (bool): default is False. If True, function returns a list with one table per channel in inputs.

    Returns:
        dict: The inputs mapped to outputs
    """
    if channel_wise:
        num_channels = inputs.shape[-1]
        io_table: list[dict] | dict = [{} for _ in range(num_channels)]
        for input, output in zip(inputs, outputs):
            for channel in range(num_channels):
                io_table[channel][tuple(input[:, channel].flatten().tolist())] = tuple(
                    output[:, channel].detach().numpy().flatten().tolist()
                )
    else:
        io_table = {}
        for input, output in zip(inputs, outputs):
            io_table[tuple(input.flatten().tolist())] = tuple(
                output.detach().numpy().flatten().tolist()
            )
    return io_table


def find_unique_elements(input_values: np.ndarray) -> np.ndarray:
    """Filters an input into its unique elements.
    Args:
        input_values (np.ndarray): The array to be filtered

    Returns:
        numpy.ndarray: A set array
    """
    output = np.unique(input_values, axis=0)
    return output


def depthwise_inputs(
    conv_layer: Module,
    kernel_size: tuple[int],
    codomain: list[int | float],
) -> np.ndarray:
    """Calculate the inputs for depthwise convolution .
    Args:
        conv_layer (Module): Layer, the convolutional layer
        codomain (list[int | float]): the codomain for the block
    Returns:
      np.ndarray: The inputs
    """
    n_channels: int = conv_layer.in_channels
    inputs = create_input_data(list(kernel_size), codomain)
    inputs = np.reshape(inputs, [-1, 1, *kernel_size])
    stacked_inputs = inputs.copy()

    if n_channels > 1:
        for _ in range(n_channels - 1):
            stacked_inputs = np.concatenate([stacked_inputs, inputs], axis=-2)

    return stacked_inputs


def _calculate_the_rank_index_that_shape_sizes_match_up_to(
    requested_shape, provided_shape
):
    rank_size_pairs = tuple(zip(reversed(provided_shape), reversed(requested_shape)))
    equalities = map(lambda x: x[0] == x[1], rank_size_pairs)
    if not all(equalities):
        raise ValueError
    return len(requested_shape) - len(rank_size_pairs)


def create_codomain_for_1d_conv(shape, codomain_elements):
    domain: torch.Tensor = torch.as_tensor(
        construct_codomain_from_elements(
            shape=shape, codomain_elements=codomain_elements, dtype="float32"
        )
    )
    return domain


def create_codomain_for_depthwise_1d_conv(shape, codomain_elements) -> Tensor:
    kernel_size = shape[1]
    channels = shape[0]
    domain: torch.Tensor = torch.as_tensor(
        construct_codomain_from_elements(
            shape=(kernel_size,), codomain_elements=codomain_elements, dtype="float32"
        )
    )
    domain = domain.reshape([-1, 1, kernel_size])
    domain = domain.repeat_interleave(channels, dim=1)
    return domain



import unittest

import torch
from torch import nn

from elasticai.creator.qat.blocks import Conv1dBlock, DepthwiseConv1dBlock, LinearBlock
from elasticai.creator.qat.layers import Binarize, Ternarize


class BlockTests(unittest.TestCase):
    def test_conv1d_block(self):
        with self.subTest("test params"):
            module = Conv1dBlock(
                in_channels=2,
                out_channels=32,
                kernel_size=3,
                conv_quantizer=Binarize(),
                activation=Ternarize(),
                stride=1,
                padding=0,
                dilation=1,
                groups=2,
                bias=True,
                padding_mode="zeros",
                constraints=None,
            )
            self.assertEqual(module.conv1d.groups, 2)
            self.assertEqual(module.batch_norm.num_features, 32)

        with self.subTest("test call"):
            module = Conv1dBlock(
                in_channels=1,
                out_channels=32,
                kernel_size=3,
                conv_quantizer=Binarize(),
                activation=Ternarize(),
                stride=1,
                padding=0,
                dilation=1,
                groups=1,
                bias=True,
                padding_mode="zeros",
                constraints=None,
            )
            module.eval()
            input = torch.randn([1, 1, 32])
            out = module(input)
            self.assertEqual(list(out.size()), [1, 32, 30])

        with self.subTest("test codomain"):
            module = Conv1dBlock(
                in_channels=1,
                out_channels=32,
                kernel_size=3,
                conv_quantizer=Binarize(),
                activation=Ternarize(),
                stride=1,
                padding=0,
                dilation=1,
                groups=1,
                bias=True,
                padding_mode="zeros",
                constraints=None,
            )
            module.eval()
            codomain = module.codomain
            self.assertEqual(codomain, Ternarize().codomain)

    def test_Depthwiseconv1d_block(self):
        with self.subTest("test params"):
            module = DepthwiseConv1dBlock(
                in_channels=4,
                out_channels=32,
                kernel_size=3,
                conv_quantizer=Binarize(),
                activation=Ternarize(),
                stride=1,
                padding=0,
                dilation=1,
                bias=True,
                padding_mode="zeros",
                constraints=None,
                pointwise_quantizer=Binarize(),
                pointwise_activation=Binarize(),
            )
            self.assertEqual(module.depthwiseConv1d.groups, 4)
            self.assertEqual(module.batchnorm.num_features, 4)
            self.assertEqual(module.pointwise_batchnorm.num_features, 32)

        with self.subTest("test call"):
            module = DepthwiseConv1dBlock(
                in_channels=2,
                out_channels=32,
                kernel_size=3,
                conv_quantizer=Binarize(),
                activation=Ternarize(),
                stride=1,
                padding=0,
                dilation=1,
                bias=True,
                padding_mode="zeros",
                constraints=None,
                pointwise_quantizer=Binarize(),
                pointwise_activation=Binarize(),
            )
            module.eval()
            input = torch.randn([1, 2, 32])
            out = module(input)
            self.assertEqual(list(out.size()), [1, 32, 30])

    def test_Linear_block(self):
        with self.subTest("test params"):
            module = LinearBlock(
                in_features=16,
                out_features=4,
                linear_quantizer=Binarize(),
                activation=Ternarize(),
                bias=True,
            )
            self.assertEqual(module.linear.out_features, 4)
            self.assertEqual(module.batchnorm.num_features, 4)

        with self.subTest("test call"):
            module = LinearBlock(
                in_features=16,
                out_features=4,
                linear_quantizer=Binarize(),
                activation=nn.Identity(),
                bias=True,
            )
            module.eval()
            input = torch.randn([1, 16])
            out = module(input)
            self.assertEqual(list(out.size()), [1, 4])


if __name__ == "__main__":
    unittest.main()



import json
from typing import Callable, Sequence

import numpy as np
import torch
from torch import Tensor

from elasticai.creator.mlframework import Module
from elasticai.creator.tags_utils import TaggedModule, get_tags, has_tag, tag

_precomputable_tag = "precomputable"

TensorLike = Callable[[], "TensorLike"] | torch.Tensor | np.ndarray


class Precomputation:
    """
    The Precomputation class provides a higher level api for precomputing the results of arbitrary pytorch modules.
    It is not a pytorch module itself nor intended to be used like one.
    Precomputations can be exported to JSON with `elasticai.creator.precomputation.JSONEncoder`.
    Note that this does not preserve the exact information about the module that was precomputed.
    There is no way currently to load an exported Precomputation object.
    """

    def __init__(self, module: Module, input_domain: TensorLike) -> None:
        super().__init__()
        self.module = module
        self.input_domain = input_domain
        self.output = None

    def _evaluate_input_domain_if_necessary(self) -> None:
        if isinstance(self.input_domain, Callable):
            self.input_domain = self.input_domain()

    @staticmethod
    def from_precomputable_tagged(module: Module):
        if not has_tag(module, _precomputable_tag):
            raise TypeError
        info_for_precomputation = get_tags(module)[_precomputable_tag]
        input_domain = info_for_precomputation["input_generator"]
        return Precomputation(module=module, input_domain=input_domain)

    def __call__(self, *args, **kwargs) -> None:
        """Precompute the input output pairs for the block handed during construction of the object.

        IMPORTANT: We assume the model is in eval mode
        """
        with torch.no_grad():
            self._evaluate_input_domain_if_necessary()
            self.output = self.module(self.input_domain)

    def __getitem__(self, item) -> Tensor:
        if item == 0:
            return self.input_domain
        elif item == 1:
            return self.output
        else:
            raise IndexError

    def dump_to(self, f):
        json.dump(JSONEncoder().default(self), f)


class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Precomputation):
            description = ""
            description += o.module.extra_repr()
            if description == "":
                description = []
                for name, child in o.module.named_children():
                    description.append(name + " " + type(child).__name__)
            return {
                "description": description,
                "shape": o.input_domain.shape[1:],
                "x": o.input_domain.numpy().tolist(),
                "y": o.output.numpy().tolist(),
            }
        else:
            return json.JSONEncoder.default(self, o)


def get_precomputations_from_direct_children(module):
    def tag_filter(child):
        return _precomputable_tag in get_tags(child)

    submodules = tuple(module.children())
    filtered_submodules = filter(tag_filter, submodules)
    yield from (
        Precomputation.from_precomputable_tagged(submodule)
        for submodule in filtered_submodules
    )


def get_precomputations_recursively(module) -> Precomputation:
    def tag_filter(submodule):
        return _precomputable_tag in get_tags(submodule)

    submodules = tuple(module.modules())
    filtered_submodules = filter(tag_filter, submodules)
    yield from (
        Precomputation.from_precomputable_tagged(submodule)
        for submodule in filtered_submodules
    )


def precomputable(
    input_shape: Sequence[int],
    input_generator: Callable[[Sequence[float | int]], np.ndarray],
) -> Callable[[Module], TaggedModule]:
    """Add all necessary information to allow later tools to precompute the specified module

    The arguments provided will be used to determine the input data that needs
    to be fed to the module to produce the precomputed input-output table.

            Example:

            model = Sequence(
                precomputable(Sequence(
                    QConv1D(...),
                    BatchNorm1D(...),
                    Binarize(),
                ), input_shape=..., input_generator=...),
                precomputable(Sequence(
                    QConv1D(...),
                    BatchNorm1D(...),
                    Binarize(),
                ), input_shape=..., input_generator=...),
                MaxPool1D(..),

            precomputations = get_precomputations(model)
            for precompute in precomputations:
                precompute()

            with open('saved_precomputations.txt', 'w') as file:
                for precomputation in precomputations:
                    file.write(precomputation)
    """
    tag_content = {
        "input_shape": input_shape,
        "input_generator": input_generator,
    }

    def decorator(module: Module) -> TaggedModule:
        return tag(module, precomputable=tag_content)

    return decorator



import math
from collections.abc import Sequence

from elasticai.creator.resource_utils import read_text
from elasticai.creator.vhdl.language import Code, hex_representation
from elasticai.creator.vhdl.number_representations import (
    FixedPoint,
    infer_total_and_frac_bits,
)


def pad_with_zeros(numbers: list[FixedPoint], target_length: int) -> list[FixedPoint]:
    zero = FixedPoint(0, *infer_total_and_frac_bits(numbers))
    return numbers + [zero] * (target_length - len(numbers))


class Rom:
    def __init__(
        self,
        rom_name: str,
        values: Sequence[FixedPoint],
        resource_option: str,
    ):
        self.rom_name = rom_name
        self.data_width, _ = infer_total_and_frac_bits(values)
        self.addr_width = self._calculate_required_addr_width_to_access_items(values)
        padded_values = pad_with_zeros(list(values), 2**self.addr_width)
        self.hex_values = list(
            map(lambda fp: hex_representation(fp.to_hex()), padded_values)
        )
        self.resource_option = resource_option

    @staticmethod
    def _calculate_required_addr_width_to_access_items(items: Sequence) -> int:
        return max(1, math.ceil(math.log2(len(items))))

    def __call__(self) -> Code:
        template = read_text("elasticai.creator.vhdl.templates", "rom.tpl.vhd")

        code = template.format(
            rom_name=self.rom_name,
            rom_addr_bitwidth=self.addr_width,
            rom_data_bitwidth=self.data_width,
            rom_value=",".join(self.hex_values),
            rom_resource_option=f'"{self.resource_option}"',
        )

        stripped_code_lines = map(str.strip, code.splitlines())

        def not_empty(line):
            return len(line) > 0

        yield from filter(not_empty, stripped_code_lines)



from functools import partial
from unittest import TestCase

from elasticai.creator.vhdl.language import (
    Architecture,
    ContextClause,
    DataType,
    Entity,
    InterfaceSignal,
    InterfaceVariable,
    LibraryClause,
    Mode,
    PortMap,
    Procedure,
    Process,
    UseClause,
)
from elasticai.creator.vhdl.number_representations import float_values_to_fixed_point
from elasticai.creator.vhdl.precomputed_scalar_function import (
    precomputed_scalar_function_process,
)


class EntityTest(TestCase):
    def test_no_name_entity(self):
        e = Entity("")
        expected = ["entity  is", "end entity ;"]
        actual = list(e())
        self.assertEqual(expected, actual)

    def test_entity_name_tanh(self):
        e = Entity("tanh")
        expected = ["entity tanh is", "end entity tanh;"]
        actual = list(e())
        self.assertEqual(expected, actual)

    def test_entity_with_generic(self):
        e = Entity("identifier")

        e.generic_list = ["test"]

        expected = [
            "entity identifier is",
            "generic (",
            "test",
            ");",
            "end entity identifier;",
        ]
        actual = list(e())
        self.assertEqual(expected, actual)

    def test_entity_with_port(self):
        e = Entity("identifier")
        e.port_list = ["test"]
        expected = [
            "entity identifier is",
            "port (",
            "test",
            ");",
            "end entity identifier;",
        ]
        actual = list(e())
        self.assertEqual(expected, actual)

    def test_entity_with_two_variables_in_generic(self):
        e = Entity("identifier")
        e.generic_list = ["first", "second"]
        expected = ["first;", "second"]
        actual = list(e())
        actual = actual[2:4]
        self.assertEqual(expected, actual)

    def test_entity_with_interface_variables(self):
        e = Entity("ident")
        e.generic_list.append(
            InterfaceVariable(
                identifier="my_var", identifier_type=DataType.INTEGER, value="16"
            )
        )
        expected = ["my_var : integer := 16"]
        actual = list(e())
        actual = actual[2:3]
        self.assertEqual(expected, actual)


class LibraryTest(TestCase):
    def test_library_with_extra_libraries(self):
        lib = ContextClause(
            library_clause=LibraryClause(logical_name_list=["ieee", "work"]),
            use_clause=UseClause(
                selected_names=[
                    "ieee.std_logic_1164.all",
                    "ieee.numeric_std.all",
                    "work.all",
                ]
            ),
        )
        expected = [
            "library ieee, work;",
            "use ieee.std_logic_1164.all;",
            "use ieee.numeric_std.all;",
            "use work.all;",
        ]
        actual = list(lib())
        self.assertEqual(expected, actual)


class ProcessTest(TestCase):
    def setUp(self) -> None:
        self.fp_list = partial(float_values_to_fixed_point, total_bits=16, frac_bits=8)

    # Note: the precomputed scalar function process is already tested, no need for trying more in- and outputs
    def test_process_empty(self):
        process = Process(
            identifier="some_name",
            input_name="some_input",
            lookup_table_generator_function=precomputed_scalar_function_process(
                x=[], y=self.fp_list([0])
            ),
        )
        expected = [
            "some_name_process: process(some_input)",
            "begin",
            'y <= "0000000000000000";',
            "end process some_name_process;",
        ]
        actual = list(process())
        self.assertEqual(expected, actual)

    def test_process_with_variables(self):
        process = Process(
            identifier="some_name",
            input_name="some_input",
            lookup_table_generator_function=precomputed_scalar_function_process(
                x=[], y=self.fp_list([0])
            ),
        )
        process.process_declaration_list = ["variable some_variable_name: integer := 0"]
        process.process_statements_list = [
            "some_variable_name := to_integer(some_variable_name)"
        ]
        expected = [
            "some_name_process: process(some_input)",
            "variable some_variable_name: integer := 0;",
            "begin",
            "some_variable_name := to_integer(some_variable_name);",
            'y <= "0000000000000000";',
            "end process some_name_process;",
        ]
        actual = list(process())
        self.assertEqual(expected, actual)


class InterfaceVariableTest(TestCase):
    def test_InterfaceVariable_empty(self):
        interface_variable = InterfaceVariable(
            identifier="some_variable", identifier_type=DataType.INTEGER
        )
        expected = ["some_variable : integer"]
        actual = list(interface_variable())
        self.assertEqual(expected, actual)

    def test_InterfaceVariable_all_parameters(self):
        interface_variable = InterfaceVariable(
            identifier="my_var",
            identifier_type=DataType.SIGNED,
            mode=Mode.IN,
            range="15 downto 0",
            value="16",
        )
        expected = ["my_var : in signed(15 downto 0) := 16"]
        actual = list(interface_variable())
        self.assertEqual(expected, actual)


class InterfaceSignalTest(TestCase):
    def test_InterfaceSignal(self):
        i = InterfaceSignal(
            identifier="y", mode=Mode.OUT, range="x", identifier_type=DataType.SIGNED
        )
        expected = ["signal y : out signed(x)"]
        actual = list(i())
        # actual = actual[2:3]
        self.assertEqual(expected, actual)


class ArchitectureTest(TestCase):
    def test_Architecture_base(self):
        a = Architecture(design_unit="z")
        expected = ["architecture rtl of z is", "begin", "end architecture rtl;"]
        actual = list(a())
        self.assertSequenceEqual(expected, actual)

    def test_Architecture_with_variables(self):
        a = Architecture(design_unit="z")
        a.architecture_declaration_list.append(
            InterfaceVariable(
                identifier="1", range="1", identifier_type=DataType.SIGNED
            )
        )
        expected = [
            "architecture rtl of z is",
            "1 : signed(1);",
            "begin",
            "end architecture rtl;",
        ]
        actual = list(a())
        self.assertSequenceEqual(expected, actual)

    def test_Architecture_with_architecture_part_as_function(self):
        def function():
            yield "some code"

        a = Architecture(design_unit="z")
        a.architecture_statement_part = function
        expected = [
            "architecture rtl of z is",
            "begin",
            "some code",
            "end architecture rtl;",
        ]
        actual = list(a())
        self.assertSequenceEqual(expected, actual)

    def test_Architecture_with_architecture_part_as_process(self):
        def function():
            def generator():
                yield "some code"

            return generator

        dummy_process = Process(
            identifier="some name",
            lookup_table_generator_function=function(),
            input_name="x",
        )
        a = Architecture(design_unit="z")
        a.architecture_statement_part = dummy_process
        expected = [
            "architecture rtl of z is",
            "begin",
            "some name_process: process(x)",
            "begin",
            "some code",
            "end process some name_process;",
            "end architecture rtl;",
        ]
        actual = list(a())
        self.assertSequenceEqual(expected, actual)

    def test_Architecture_with_assignment(self):
        a = Architecture(design_unit="z")
        a.architecture_assignment_list.append("A <= B")
        expected = [
            "architecture rtl of z is",
            "begin",
            "A <= B;",
            "end architecture rtl;",
        ]
        actual = list(a())
        self.assertSequenceEqual(expected, actual)

    def test_Architecture_with_port_map(self):
        architecture = Architecture(design_unit="lstm_cell")
        port_map = PortMap(map_name="something", component_name="sigmoid")
        port_map.signal_list.append("x => y")
        port_map.signal_list.append("a => b")
        port_map.signal_list.append("c => d")
        architecture.architecture_port_map_list.append(port_map)

        expected = [
            "architecture rtl of lstm_cell is",
            "begin",
            "something: sigmoid",
            "port map (",
            "x => y,",
            "a => b,",
            "c => d",
            ");",
            "end architecture rtl;",
        ]
        actual = list(architecture())
        self.assertSequenceEqual(expected, actual)

    def test_Architecture_with_process_statements_list(self):
        dummy_process = Process(
            identifier="H_OUT",
            input_name="o,c_new",
        )
        dummy_process.process_statements_list.append(
            "h_new <= shift_right((o*c_new), FRAC_WIDTH)(DATA_WIDTH-1 downto 0)"
        )
        a = Architecture(design_unit="z")
        a.architecture_statement_part = dummy_process
        expected = [
            "architecture rtl of z is",
            "begin",
            "H_OUT_process: process(o,c_new)",
            "begin",
            "h_new <= shift_right((o*c_new), FRAC_WIDTH)(DATA_WIDTH-1 downto 0);",
            "end process H_OUT_process;",
            "end architecture rtl;",
        ]
        actual = list(a())
        self.assertSequenceEqual(expected, actual)


class PortMapTest(TestCase):
    def test_PortMap(self):
        port_map = PortMap(map_name="something", component_name="lstm")
        port_map.signal_list.append("x => y")
        expected = ["something: lstm", "port map (", "x => y", ");"]
        actual = list(port_map())
        self.assertSequenceEqual(expected, actual)


class ProcedureTest(TestCase):
    def test_procedure(self):
        procedure = Procedure(identifier="abc")
        expected = ["procedure abc (", "begin", "end abc;"]
        actual = list(procedure())
        self.assertSequenceEqual(expected, actual)

    def test_procedure_with_declaration_and_statement_list(self):
        procedure = Procedure(identifier="abc")
        procedure.declaration_list = [
            "some_variable : in some_name(some_parameter)",
        ]
        procedure.declaration_list_with_is = [
            "signal some_variable_1 : out some_name(some_parameter))"
        ]
        procedure.statement_list = [
            "xyz <= efg",
        ]
        expected = [
            "procedure abc (",
            "some_variable : in some_name(some_parameter);",
            "signal some_variable_1 : out some_name(some_parameter)) is",
            "begin",
            "xyz <= efg;",
            "end abc;",
        ]
        actual = list(procedure())
        self.assertSequenceEqual(expected, actual)



from functools import partial
from typing import Optional, Tuple
from unittest import SkipTest
from unittest.mock import patch

import torch
from torch import Tensor

from elasticai.creator.qat.constraints import WeightClipper
from elasticai.creator.qat.layers import (
    QLSTM,
    BatchNormedActivatedConv1d,
    Binarize,
    BinaryActivatedConv1d,
    BinarySplitConv,
    ChannelShuffle,
    MultilevelResidualBinarizationConv1d,
    QConv1d,
    QConv2d,
    QLinear,
    QLSTMCell,
    QuantizeTwoBit,
    ResidualQuantization,
    SplitConvolutionBase,
    Ternarize,
    TernaryConvolution1d,
    TernarySplitConv,
    TwoBitSplitConv,
)
from elasticai.creator.tests.tensor_test_case import TensorTestCase


class MockQLSTMCell:
    def __init__(*args, **kwargs):
        pass

    # it is the forward function in a real LSTM cell
    def __call__(self, input: Tensor, hx: Optional[Tuple[Tensor, Tensor]] = None):
        return hx


class LayerTests(TensorTestCase):
    def test_ResidualQuantization(self):
        with self.subTest("test call with different values 1,-1,0"):
            layer = ResidualQuantization()
            out = layer(torch.unsqueeze(torch.Tensor([1, -1, 0]), dim=0))
            self.assertEqual(out.tolist(), [[1.0, -1.0, 1.0, 1.0, -1.0, -1.0]])

        with self.subTest("test if Backward pass requires grad"):
            layer = ResidualQuantization()
            out = layer(torch.unsqueeze(torch.Tensor([2]), 0))
            self.assertEqual(out.requires_grad, True)
        with self.subTest("thresholds"):
            layer = ResidualQuantization()
            threshold = layer.thresholds
            self.assertEqual(layer.thresholds.tolist(), [0, 0.5])
        with self.subTest("codomain"):
            layer = ResidualQuantization()
            self.assertSequenceEqual(
                layer.codomain, [(x, y) for x in [-1, 1] for y in [-1, 1]]
            )

    def test_Ternarize(self):
        with self.subTest("test call with different values 1,-1,0"):
            layer = Ternarize()
            out = layer(torch.Tensor([1, -1, 0]))
            self.assertEqual(out.tolist(), [1, -1, 0])

        with self.subTest("test does not require grad with default init"):
            # shouldn't require grad so no backward call
            layer = Ternarize()
            out = layer(torch.Tensor([2]))
            self.assertEqual(out.requires_grad, False)

        with self.subTest("thresholds"):
            layer = Ternarize(0.25)
            self.assertEqual(layer.thresholds.tolist(), [0.25, -0.25])

        with self.subTest("Check if widening is trainable when option True"):
            layer = Ternarize(zero_window_width=0.5, trainable=True)
            out = layer(torch.Tensor([2]))
            self.assertEqual(out.requires_grad, True)

    def test_QuantizeTwoBit(self):
        with self.subTest("test call with  0.3 factor and 2 as ipnut"):
            layer = QuantizeTwoBit(0.3)
            out = layer(torch.unsqueeze(torch.Tensor([2]), 0))
            self.assertEqual(out.tolist(), [[1, 1.0]])
        with self.subTest("test factors are trainable"):
            layer = QuantizeTwoBit(0.3)
            out = layer(torch.unsqueeze(torch.Tensor([2]), 0))
            self.assertEqual(out.requires_grad, True)
        with self.subTest("Check thresholds value with factors 0.3"):
            layer = QuantizeTwoBit(0.3)
            thresholds = layer.thresholds.tolist()
            self.assertEqual(
                [0.0, round(thresholds[1], 2)], [0.0, 0.3]
            )  # precision loss

    def test_Qlinear(self):
        with self.subTest("test call if binarized and has bias"):
            layer = QLinear(1, 2, Binarize())
            layer.weight = torch.Tensor([[2], [-2]])
            layer.bias = torch.Tensor([[3, -3]])
            input = torch.unsqueeze(torch.Tensor([3.0]), 0)
            out = layer(input)
            self.assertEqual(out.tolist(), [[4.0, -4.0]])

        with self.subTest("test Backward pass changes the weight gradient"):
            layer = QLinear(1, 2, Binarize())
            layer.weight = torch.Tensor([[0.6], [0.7]])
            old_grad = layer.parametrizations.weight.original.grad
            layer.bias = torch.Tensor([[3.0, -3.0]])
            input = torch.unsqueeze(torch.Tensor([3.0]), 0)
            out = layer(input)
            loss = (out - torch.Tensor([-1.0])).sum()
            loss.backward()
            new_weight_grad = layer.parametrizations.weight.original.grad
            self.assertNotEqual(new_weight_grad[0], old_grad)
            self.assertNotEqual(new_weight_grad[1], old_grad)

        with self.subTest("test weight clipper constraint QLinear"):
            layer = QLinear(
                in_features=1,
                out_features=2,
                quantizer=Binarize(),
                constraints=[WeightClipper()],
            )
            layer.weight = torch.Tensor([[-2], [2]])
            layer.bias = torch.Tensor([[2, 2]])
            input = torch.unsqueeze(torch.Tensor([3.0]), 0)
            # clip to [-1, 1]
            with torch.no_grad():
                layer.apply_constraint()
            out = layer(input)
            # 3*(-1)+2=-1 and 3*1+2=5
            self.assertEqual(out.tolist(), [[-2, 4]])

        with self.subTest("test throw error if quanitzer is None"):
            self.assertRaises(TypeError, QLinear, 1, 1, None)

    def test_QConv1d(self):
        with self.subTest("test call with binarizer and bias"):
            layer = QConv1d(1, 1, (2,), Binarize())
            layer.weight = torch.zeros(1, 1, 2)
            layer.bias = torch.Tensor([3.0])
            input = torch.unsqueeze(
                torch.Tensor(
                    [[3.0, 2]],
                ),
                0,
            )
            out = layer(input)
            self.assertTensorEquals(out, torch.tensor([[[6.0]]]))

        with self.subTest("test Backward pass changes the weight gradient"):
            layer = QConv1d(1, 1, 2, Binarize())
            layer.weight = torch.zeros(1, 1, 2)
            layer.bias = torch.Tensor([3.0])
            old_grad = layer.parametrizations.weight.original.grad
            input = torch.unsqueeze(
                torch.Tensor(
                    [[3.0, 2]],
                ),
                0,
            )
            out = layer(input)
            loss = (out - torch.Tensor([-1.0])).sum()
            loss.backward()
            new_weight_grad = layer.parametrizations.weight.original.grad
            self.assertNotEqual(new_weight_grad[0], old_grad)

        with self.subTest("test constraint QConv1d"):
            layer = QConv1d(
                in_channels=1,
                out_channels=1,
                kernel_size=2,
                quantizer=Binarize(),
                constraints=[WeightClipper()],
            )
            layer.weight = torch.tensor([[[-2.0, 2.0]]])
            layer.bias = torch.Tensor([3.0])
            input = torch.unsqueeze(
                torch.Tensor(
                    [[3.0, 2.0]],
                ),
                0,
            )
            # clip to [-1, 1]
            with torch.no_grad():
                layer.apply_constraint()
            out = layer(input)
            # (3*(-1)+2*1)+3 = 2
            self.assertEqual(out.tolist(), [[[0.0]]])

        with self.subTest("test apply constraint QConv1d if constaint is none"):
            layer = QConv1d(
                in_channels=1, out_channels=1, kernel_size=2, quantizer=Binarize()
            )
            layer.weight = torch.tensor([[[-2.0, 2.0]]])
            layer.bias = torch.Tensor([3.0])
            input = torch.unsqueeze(
                torch.Tensor(
                    [[3.0, 2.0]],
                ),
                0,
            )
            # clip to [-1, 1]
            with torch.no_grad():
                layer.apply_constraint()
            out = layer(input)
            # (3*(-1)+2*1)+3 = 2
            self.assertEqual(out.tolist(), [[[0.0]]])

        with self.subTest("test throw error if quanitzer is None"):
            self.assertRaises(TypeError, QConv1d, 1, 1, 2, None)

    def test_QConv2d(self):
        with self.subTest("test call with binarization and bias"):
            layer = QConv2d(1, 1, 2, Binarize())
            layer.weight = torch.zeros(1, 1, 2, 2)
            layer.bias = torch.Tensor([3.0])
            input = torch.unsqueeze(
                torch.Tensor(
                    [[[3.0, 2.0], [3.0, 2.0]]],
                ),
                0,
            )
            out = layer(input)
            self.assertEqual(out.tolist(), [[[[11.0]]]])

        with self.subTest("test Backward pass, changed weights"):
            layer = QConv2d(1, 1, 2, Binarize())
            layer.weight = torch.zeros(1, 1, 2, 2)
            layer.bias = torch.Tensor([3.0])
            input = torch.unsqueeze(
                torch.Tensor(
                    [[[3.0, 2.0], [3.0, 2.0]]],
                ),
                0,
            )
            old_grad = layer.parametrizations.weight.original.grad
            out = layer(input)
            loss = (out - torch.Tensor([-1.0])).sum()
            loss.backward()
            new_weight_grad = layer.parametrizations.weight.original.grad
            self.assertNotEqual(new_weight_grad[0], old_grad)

        with self.subTest("test weight clipper constraint QConv2d"):
            layer = QConv2d(
                in_channels=1,
                out_channels=1,
                kernel_size=2,
                quantizer=Binarize(),
                constraints=[WeightClipper()],
            )
            layer.weight = torch.tensor([[[[-2.0, 2.0], [0.0, 1.0]]]])
            layer.bias = torch.Tensor([3.0])
            input = torch.unsqueeze(
                torch.Tensor(
                    [[[3.0, 2.0], [3.0, 2.0]]],
                ),
                0,
            )
            # clip to [-1, 1]
            with torch.no_grad():
                layer.apply_constraint()
            out = layer(input)
            # (3*(-1)+2*1+3*0+2*1)+3 = 4
            self.assertEqual(out.tolist(), [[[[5.0]]]])

        with self.subTest("test throw error if quanitzer is None"):
            self.assertRaises(TypeError, QConv2d, 1, 1, 2, None)

    def test_BatchNormedActivatedConv(self):
        layer = BatchNormedActivatedConv1d(
            in_channels=1,
            out_channels=2,
            kernel_size=2,
            groups=1,
            bias=False,
            activation=Binarize,
            channel_multiplexing_factor=1,
        )
        layer.conv.weight = torch.nn.Parameter(torch.ones_like(layer.conv.weight))
        test_input = torch.ones((2, 1, 3))
        output = layer(test_input)
        expected = torch.ones(2, 2, 2)
        self.assertTrue(torch.all((expected == output)))

    def test_TernaryConvolution(self):
        layer = TernaryConvolution1d(
            in_channels=1, out_channels=2, kernel_size=2, groups=1, bias=False
        )
        self.assertEqual(type(layer.quantize), type(Ternarize()))
        self.assertEqual(layer.channel_multiplexing_factor, 1)

    def test_MultilevelResidualBinarizationConv(self):
        layer = MultilevelResidualBinarizationConv1d(
            in_channels=1, out_channels=2, kernel_size=2, groups=1, bias=False
        )
        self.assertEqual(type(layer.quantize), type(QuantizeTwoBit()))
        self.assertEqual(layer.channel_multiplexing_factor, 2)

    def test_BinaryActivatedConv(self):
        layer = BinaryActivatedConv1d(
            in_channels=1, out_channels=2, kernel_size=2, groups=1, bias=False
        )
        self.assertEqual(type(layer.quantize), type(Binarize()))
        self.assertEqual(layer.channel_multiplexing_factor, 1)

    def test_SplitConvolutionBase(self):
        layer_function = partial(
            BatchNormedActivatedConv1d,
            activation=Binarize,
            channel_multiplexing_factor=1,
        )
        layer = SplitConvolutionBase(
            in_channels=2,
            out_channels=4,
            kernel_size=2,
            convolution=layer_function,
            codomain_elements=[-1, 1],
        )
        layer.depthwise.conv.weight = torch.nn.Parameter(
            torch.ones_like(layer.depthwise.conv.weight)
        )
        layer.pointwise.conv.weight = torch.nn.Parameter(
            torch.ones_like(layer.pointwise.conv.weight)
        )
        test_input = torch.ones((2, 2, 3))
        output = layer(test_input)
        expected = torch.ones(2, 4, 2)
        self.assertTrue(torch.all((expected == output)))
        self.assertEqual(layer.depthwise.conv.groups, 2)

    def test_BinarySplitConv(self):
        layer = BinarySplitConv(in_channels=1, out_channels=2, kernel_size=2)
        self.assertEqual(type(layer.depthwise.quantize), type(Binarize()))
        self.assertEqual(layer.depthwise.channel_multiplexing_factor, 1)
        self.assertEqual(type(layer.pointwise.quantize), type(Binarize()))

    def test_TernarySplitConv(self):
        layer = TernarySplitConv(in_channels=1, out_channels=2, kernel_size=2)
        self.assertEqual(type(layer.depthwise.quantize), type(Ternarize()))
        self.assertEqual(layer.depthwise.channel_multiplexing_factor, 1)
        self.assertEqual(type(layer.pointwise.quantize), type(Ternarize()))

    def test_TwoBitSplitConv(self):
        layer = TwoBitSplitConv(in_channels=1, out_channels=2, kernel_size=2)
        self.assertEqual(type(layer.depthwise.quantize), type(QuantizeTwoBit()))
        self.assertEqual(layer.depthwise.channel_multiplexing_factor, 2)
        self.assertEqual(type(layer.pointwise.quantize), type(QuantizeTwoBit()))

    @SkipTest
    def test_QLSTMCell(self):
        with self.subTest("full res QLSTM cell equal PyTorch LSTM cell (with bias)"):
            lstm_cell = torch.nn.LSTMCell(input_size=3, hidden_size=5, bias=True)

            def identity(x):
                return x

            qlstm_cell = QLSTMCell(
                input_size=3,
                hidden_size=5,
                bias=True,
                weight_quantizer=identity,
                state_quantizer=lambda x: x,
            )
            qlstm_cell.weight_ih = lstm_cell.weight_ih
            qlstm_cell.weight_hh = lstm_cell.weight_hh
            qlstm_cell.bias_ih = lstm_cell.bias_ih
            qlstm_cell.bias_hh = lstm_cell.bias_hh
            inp = torch.rand(1, 3)
            lstm_h1, lstm_c1 = lstm_cell(inp)
            qlstm_h1, qlstm_c1 = qlstm_cell(inp)
            self.assertTrue(torch.equal(lstm_h1, qlstm_h1))
            self.assertTrue(torch.equal(lstm_c1, qlstm_c1))

        with self.subTest("full res QLSTM cell equal PyTorch LSTM cell (without bias)"):
            lstm_cell = torch.nn.LSTMCell(input_size=3, hidden_size=5, bias=False)
            qlstm_cell = QLSTMCell(
                input_size=3,
                hidden_size=5,
                bias=False,
            )
            qlstm_cell.weight_ih = lstm_cell.weight_ih
            qlstm_cell.weight_hh = lstm_cell.weight_hh
            qlstm_cell.bias_ih = lstm_cell.bias_ih
            qlstm_cell.bias_hh = lstm_cell.bias_hh
            inp = torch.rand(1, 3)
            lstm_h1, lstm_c1 = lstm_cell(inp)
            qlstm_h1, qlstm_c1 = qlstm_cell(inp)
            self.assertTrue(torch.equal(lstm_h1, qlstm_h1))
            self.assertTrue(torch.equal(lstm_c1, qlstm_c1))

        with self.subTest("test binarized cell states and weights"):
            cell = QLSTMCell(
                input_size=1,
                hidden_size=1,
                bias=True,
                state_quantizer=Binarize(),
                weight_quantizer=Binarize(),
            )
            cell.weight_ih = torch.nn.Parameter(torch.ones_like(cell.weight_ih))
            cell.weight_hh = torch.nn.Parameter(torch.ones_like(cell.weight_hh) * (-1))
            cell.bias_ih = torch.nn.Parameter(torch.ones_like(cell.bias_ih))
            cell.bias_hh = torch.nn.Parameter(torch.ones_like(cell.bias_hh) * (-1))
            inp = torch.as_tensor([[1.0]])
            actual_outputs = cell(inp)
            actual_outputs = tuple(
                map(
                    lambda output: [
                        [round(n, 4) for n in row] for row in output.tolist()
                    ],
                    actual_outputs,
                )
            )
            target_outputs = ([[0.2311]], [[0.5]])
            self.assertListEqual(actual_outputs[0], target_outputs[0])
            self.assertListEqual(actual_outputs[1], target_outputs[1])

    def test_QLSTM(self):
        lstm_cell_id = "elasticai.creator.qat.layers.QLSTMCell"
        qlstm_parameters = {
            "input_size": 0,
            "hidden_size": 1,
            "state_quantizer": lambda x: x,
            "weight_quantizer": lambda x: x,
            "bias": False,
        }
        with patch(lstm_cell_id, new=MockQLSTMCell) as _:
            with self.subTest("test QLSTM default"):
                layer = QLSTM(**qlstm_parameters)
                input = torch.Tensor([[1.0, 1.0], [1.0, 1.0]])
                state = [torch.Tensor([1.0]), torch.Tensor([1.0])]
                outputs, cell_state = layer(input, state)
                self.assertEqual(outputs.tolist(), [[1.0], [1.0]])
                self.assertEqual(cell_state[0].tolist(), [1.0])
                self.assertEqual(cell_state[1].tolist(), [1.0])

        with patch(lstm_cell_id, new=MockQLSTMCell) as _:
            with self.subTest("test QLSTM other input + state"):
                layer = QLSTM(**qlstm_parameters)
                input = torch.Tensor([[1.0, 1.0], [1.0, 1.0], [1.0, 1.0]])
                state = [torch.Tensor([5.0]), torch.Tensor([-1.0])]
                outputs, cell_state = layer(input, state)
                self.assertEqual(outputs.tolist(), [[5.0], [5.0], [5.0]])
                self.assertEqual(cell_state[0].tolist(), [5.0])
                self.assertEqual(cell_state[1].tolist(), [-1.0])

    def test_shuffle_block_input_and_output_same_with_1_group(self):
        layer = ChannelShuffle(groups=1)
        input = torch.rand((2, 2, 3, 2))
        output = layer(input)
        n = torch.eq(input, output)
        self.assertTrue(torch.all(torch.eq(input, output)).item())

    def test_shuffle_block_input_and_output_with_2_group(self):
        layer = ChannelShuffle(groups=2)
        input = torch.tensor(
            [
                [[1], [2], [3], [4]],
            ]
        )
        output = layer(input)
        expected = torch.tensor(
            [
                [[1], [3], [2], [4]],
            ]
        )
        self.assertTrue(torch.all(torch.eq(output, expected)).item())



""" This file corresponds to the functional module in pytorch. We keep here custom autograd functions
"""

from typing import Any

import torch
from torch import jit


@jit.script
def _heaviside(x):
    return torch.heaviside(x, torch.tensor([1.0]))


# noinspection PyPep8Naming,PyAbstractClass
class heaviside_with_ste(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> Any:
        if len(args) == 0:
            raise TypeError
        x = args[0]
        masked_zero_for_out_of_range = _heaviside(x + 1) - _heaviside(x - 1)
        ctx.save_for_backward(masked_zero_for_out_of_range)
        y = 2 * _heaviside(x) - 1
        return y

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        (masked_zero_for_out_of_range_inputs,) = ctx.saved_tensors
        return grad_outputs[0] * masked_zero_for_out_of_range_inputs


binarize = heaviside_with_ste



from functools import partial
from unittest import TestCase

from elasticai.creator.vhdl.language import hex_representation
from elasticai.creator.vhdl.lstm_testbench_generator import TestCasesLSTMCommonGate
from elasticai.creator.vhdl.number_representations import (
    FixedPoint,
    float_values_to_fixed_point,
)
from elasticai.creator.vhdl.precomputed_scalar_function import (
    TestCasesPrecomputedScalarFunction,
)


class TestCasesLSTMCommonGateTest(TestCase):
    def test_TestCasesLSTMCommonGate(self):
        to_fp = partial(FixedPoint, total_bits=8, frac_bits=0)

        def to_hex(value: FixedPoint) -> str:
            return hex_representation(value.to_hex())

        x_mem_list_for_testing = [to_fp(x) for x in range(0, 3)]
        w_mem_list_for_testing = [to_fp(x) for x in range(10, 13)]
        b_list_for_testing = [to_fp(x) for x in range(20, 23)]
        y_list_for_testing = [to_fp(x) for x in range(30, 33)]
        y_variable_name = "out"
        test_cases_lstm_common_gate = TestCasesLSTMCommonGate(
            x_mem_list_for_testing=x_mem_list_for_testing,
            w_mem_list_for_testing=w_mem_list_for_testing,
            b_list_for_testing=b_list_for_testing,
            y_list_for_testing=y_list_for_testing,
            y_variable_name=y_variable_name,
        )
        expected = [
            f'report "======Simulation Start======" severity Note',
            f"vector_len <= to_unsigned(10, VECTOR_LEN_WIDTH)",
            f"X_MEM <= ({to_hex(x_mem_list_for_testing[0])})",
            f"W_MEM <= ({to_hex(w_mem_list_for_testing[0])})",
            f"b <= {to_hex(b_list_for_testing[0])}",
            f"reset <= '1'",
            f"wait for 2*clk_period",
            f"wait until clock = '0'",
            f"reset <= '0'",
            f"wait until ready = '1'",
            f"report \"expected output is {y_list_for_testing[0]}, value of '{y_variable_name}' is \" & integer'image(to_integer(signed({y_variable_name})))",
            f'assert {y_variable_name}={y_list_for_testing[0]} report "The 0. test case fail" severity error',
            f"reset <= '1'",
            f"wait for 1*clk_period",
            f"X_MEM <= ({to_hex(x_mem_list_for_testing[1])})",
            f"W_MEM <= ({to_hex(w_mem_list_for_testing[1])})",
            f"b <= {to_hex(b_list_for_testing[1])}",
            f"reset <= '1'",
            f"wait for 2*clk_period",
            f"wait until clock = '0'",
            f"reset <= '0'",
            f"wait until ready = '1'",
            f"report \"expected output is {y_list_for_testing[1]}, value of '{y_variable_name}' is \" & integer'image(to_integer(signed({y_variable_name})))",
            f'assert {y_variable_name}={y_list_for_testing[1]} report "The 1. test case fail" severity error',
            f"reset <= '1'",
            f"wait for 1*clk_period",
            f"X_MEM <= ({to_hex(x_mem_list_for_testing[2])})",
            f"W_MEM <= ({to_hex(w_mem_list_for_testing[2])})",
            f"b <= {to_hex(b_list_for_testing[2])}",
            f"reset <= '1'",
            f"wait for 2*clk_period",
            f"wait until clock = '0'",
            f"reset <= '0'",
            f"wait until ready = '1'",
            f"report \"expected output is {y_list_for_testing[2]}, value of '{y_variable_name}' is \" & integer'image(to_integer(signed({y_variable_name})))",
            f'assert {y_variable_name}={y_list_for_testing[2]} report "The 2. test case fail" severity error',
            f"reset <= '1'",
            f"wait for 1*clk_period",
            f'report "======Simulation Success======" severity Note',
            f'report "Please check the output message." severity Note',
            f"wait",
        ]
        actual = list(test_cases_lstm_common_gate())
        self.assertSequenceEqual(expected, actual)

    def test_TestCasesLSTMCommonGate_different_lengths_of_lists(self):
        x_mem_list_for_testing = [
            "some_string_00",
            "some_string_01",
        ]
        w_mem_list_for_testing = [
            "some_string_10",
            "some_string_11",
            "some_string_12",
            "some_string_13",
        ]
        b_list_for_testing = ["some_string_20", "some_string_21", "some_string_22"]
        y_list_for_testing = [1, 2, 3]
        y_variable_name = "out"
        self.assertRaises(
            AssertionError,
            TestCasesLSTMCommonGate,
            x_mem_list_for_testing,
            w_mem_list_for_testing,
            b_list_for_testing,
            y_list_for_testing,
            y_variable_name,
        )


class TestCasesPrecomputedScalarFunctionTest(TestCase):
    def setUp(self) -> None:
        self.data_width, self.frac_width = 8, 0
        self.fp_list = partial(
            float_values_to_fixed_point,
            total_bits=self.data_width,
            frac_bits=self.frac_width,
        )

    def test_TestCasesPrecomputedScalarFunction(self):
        x_list_for_testing = [1, 2, 3]
        y_list_for_testing = [-5, -2, -3]
        x_variable_name = "x_name"
        y_variable_name = "y_name"
        test_cases_precomputed_scalar_function = TestCasesPrecomputedScalarFunction(
            x_list_for_testing=self.fp_list(x_list_for_testing),
            y_list_for_testing=self.fp_list(y_list_for_testing),
            x_variable_name=x_variable_name,
            y_variable_name=y_variable_name,
        )
        expected = [
            f'report "======Simulation Start======" severity Note',
            f"{x_variable_name} <= to_signed({x_list_for_testing[0]},{self.data_width})",
            f"wait for 1*clk_period",
            f"report \"The value of '{y_variable_name}' is \" & integer'image(to_integer(unsigned({y_variable_name})))",
            f'assert {y_variable_name}={y_list_for_testing[0]} report "The test case {x_list_for_testing[0]} fail" severity failure',
            f"{x_variable_name} <= to_signed({x_list_for_testing[1]},{self.data_width})",
            f"wait for 1*clk_period",
            f"report \"The value of '{y_variable_name}' is \" & integer'image(to_integer(unsigned({y_variable_name})))",
            f'assert {y_variable_name}={y_list_for_testing[1]} report "The test case {x_list_for_testing[1]} fail" severity failure',
            f"{x_variable_name} <= to_signed({x_list_for_testing[2]},{self.data_width})",
            f"wait for 1*clk_period",
            f"report \"The value of '{y_variable_name}' is \" & integer'image(to_integer(unsigned({y_variable_name})))",
            f'assert {y_variable_name}={y_list_for_testing[2]} report "The test case {x_list_for_testing[2]} fail" severity failure',
            f'report "======Simulation Success======" severity Note',
            f'report "Please check the output message." severity Note',
            f"wait",
        ]
        actual = list(test_cases_precomputed_scalar_function())
        self.assertSequenceEqual(expected, actual)

    def test_TestCasesPrecomputedScalarFunction_different_lengths_of_list(self):
        x_list_for_testing = [1, 2, 3]
        y_list_for_testing = [-5, -3]
        x_variable_name = "x_name"
        y_variable_name = "y_name"
        self.assertRaises(
            AssertionError,
            TestCasesPrecomputedScalarFunction,
            self.fp_list(x_list_for_testing),
            self.fp_list(y_list_for_testing),
            x_variable_name,
            y_variable_name,
        )



import torch
from torch.nn import Parameter


def random_mask_4d(
    out_channels: int,
    kernel_size: int | tuple,
    in_channels: int,
    groups: int,
    params_per_channel: int,
):
    """
    Creates a 4d mask with a  number of nonzero elements per out channels (index 0) equals to the params_per_channel randomly selected
    Args:
        out_channels:
        kernel_size:
        in_channels:
        groups:
        params_per_channel:

    Returns:

    """
    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)
    mask = Parameter(
        torch.zeros((out_channels, in_channels // groups, *kernel_size)),
        requires_grad=False,
    )
    for i in range(out_channels):
        original_shape = mask.shape[1:]
        flattened_channel = mask[i].view(-1)
        random_indices = torch.randperm(flattened_channel.shape[0])
        flattened_channel[random_indices[:params_per_channel]] = 1
        mask[i] = torch.reshape(flattened_channel, original_shape)
    return mask


def fixed_offset_mask_4d(
    out_channels: int,
    kernel_size: int | tuple,
    in_channels: int,
    groups: int,
    axis_width: int,
    offset_axis=1,
):
    """
    Creates a 4d mask with an offset per out channel, on each channel each element of the offset indices part of the
    offset axis is set to 1. Can select more than 1 by setting an axis width.
    The offset will wrap over the axis so index % len(axis)

    Args:
        out_channels:
        kernel_size:
        in_channels:
        groups:
        axis_width:
        offset_axis:

    Returns:

    """
    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)
    mask = Parameter(
        torch.zeros((out_channels, in_channels // groups, *kernel_size)),
        requires_grad=False,
    )

    for i in range(out_channels):

        axis_group_index = i % (mask.size()[offset_axis] // axis_width)
        axis_indices = list(
            map(lambda x: x + axis_group_index * axis_width, list(range(axis_width)))
        )
        if offset_axis == 1:
            mask[i, axis_indices, :, :] = 1
        if offset_axis == 2:
            mask[i, :, axis_indices, :] = 1
        if offset_axis == 3:
            mask[i, :, :, axis_indices] = 1

    return mask


class FixedOffsetMask4d(torch.nn.Module):
    def __init__(
        self,
        out_channels: int,
        kernel_size: int | tuple,
        in_channels: int,
        groups: int,
        axis_width: int,
        offset_axis=1,
    ):
        super().__init__()
        self.mask = fixed_offset_mask_4d(
            out_channels=out_channels,
            kernel_size=kernel_size,
            in_channels=in_channels,
            groups=groups,
            axis_width=axis_width,
            offset_axis=offset_axis,
        )

    def forward(self, input_values):
        return input_values * self.mask



from functools import partial
from unittest import TestCase

from elasticai.creator.vhdl.number_representations import (
    FixedPoint,
    ToLogicEncoder,
    float_values_to_fixed_point,
    infer_total_and_frac_bits,
)


class FixedPointTest(TestCase):
    def test_zero(self):
        fp_value = FixedPoint(0, total_bits=1, frac_bits=0)
        self.assertEqual(0, int(fp_value))

    def test_1_with_8_total_bits_3_frac_bits(self):
        fp_value = FixedPoint(1, total_bits=8, frac_bits=3)
        self.assertEqual(8, int(fp_value))

    def test_minus_1_with_8_total_bits_3_frac_bits(self):
        fp_value = FixedPoint(-1, total_bits=8, frac_bits=3)
        self.assertEqual(248, int(fp_value))

    def test_minus_3_21_with_16_total_bits_12_frac_bits(self):
        fp_value = FixedPoint(-3.21, total_bits=16, frac_bits=12)
        self.assertEqual(52388, int(fp_value))

    def test_initialize_with_value_not_in_range(self):
        with self.assertRaises(ValueError):
            _ = FixedPoint(100, total_bits=4, frac_bits=2)

    def test_fixed_point_to_float_5_36(self):
        fp_value = FixedPoint(5.36, total_bits=12, frac_bits=6)
        self.assertAlmostEqual(5.36, float(fp_value), places=2)

    def test_fixed_point_to_float_minus_5_36(self):
        fp_value = FixedPoint(-5.36, total_bits=16, frac_bits=12)
        self.assertAlmostEqual(-5.36, float(fp_value), places=2)

    def test_to_hex_zero_with_one_bits(self):
        fp_value = FixedPoint(0, total_bits=1, frac_bits=0)
        self.assertEqual("0", fp_value.to_hex())

    def test_to_hex_zero_with_six_bits(self):
        fp_value = FixedPoint(0, total_bits=6, frac_bits=0)
        self.assertEqual("00", fp_value.to_hex())

    def test_to_hex_zero_with_sixteen_bits(self):
        fp_value = FixedPoint(0, total_bits=16, frac_bits=0)
        self.assertEqual("0000", fp_value.to_hex())

    def test_to_hex_minus_one_with_sixteen_bits(self):
        fp_value = FixedPoint(-1, total_bits=16, frac_bits=0)
        self.assertEqual("ffff", fp_value.to_hex())

    def test_to_hex_minus_three_with_three_bits(self):
        fp_value = FixedPoint(-3, total_bits=3, frac_bits=0)
        self.assertEqual("5", fp_value.to_hex())

    def test_to_hex_minus_254_with_sixteen_bits(self):
        fp_value = FixedPoint(-254, total_bits=16, frac_bits=0)
        self.assertEqual("ff02", fp_value.to_hex())

    def test_to_hex_minus_19_5_with_16_bits(self):
        fp_value = FixedPoint(-19.5, total_bits=16, frac_bits=8)
        self.assertEqual("ec80", fp_value.to_hex())

    def test_to_bin_zero_with_one_bits(self):
        fp_value = FixedPoint(0, total_bits=1, frac_bits=0)
        self.assertEqual("0", fp_value.to_bin())

    def test_to_bin_zero_with_three_bits(self):
        fp_value = FixedPoint(0, total_bits=3, frac_bits=0)
        self.assertEqual("000", fp_value.to_bin())

    def test_to_bin_five_with_four_bits(self):
        fp_value = FixedPoint(5, total_bits=4, frac_bits=0)
        self.assertEqual("0101", fp_value.to_bin())

    def test_to_bin_minus_one_with_two_bits(self):
        fp_value = FixedPoint(-1, total_bits=2, frac_bits=0)
        self.assertEqual("11", fp_value.to_bin())

    def test_to_bin_minus_two_with_two_bits(self):
        fp_value = FixedPoint(-2, total_bits=2, frac_bits=0)
        self.assertEqual("10", fp_value.to_bin())

    def test_to_bin_minus_256_with_sixteen_bits(self):
        fp_value = FixedPoint(-256, total_bits=16, frac_bits=0)
        self.assertEqual("1111111100000000", fp_value.to_bin())

    def test_to_bin_minus_254_with_sixteen_bits(self):
        fp_value = FixedPoint(-254, total_bits=16, frac_bits=0)
        self.assertEqual("1111111100000010", fp_value.to_bin())

    def test_to_bin_minus_19_5_with_16_bits(self):
        fp_value = FixedPoint(-19.5, total_bits=16, frac_bits=8)
        self.assertEqual("1110110010000000", fp_value.to_bin())

    def test_from_int(self):
        fp_value = FixedPoint.from_int(52388, total_bits=16, frac_bits=12)
        self.assertAlmostEqual(-3.21, float(fp_value), places=2)

    def test_repr(self):
        fp_value = FixedPoint(value=3.2, total_bits=8, frac_bits=4)
        target_repr = "FixedPoint(value=3.2, total_bits=8, frac_bits=4)"
        self.assertEqual(target_repr, repr(fp_value))

    def test_str(self):
        fp_value = FixedPoint(5, total_bits=8, frac_bits=4)
        target_str = "80"
        self.assertEqual(target_str, str(fp_value))

    def test_bin_iter(self):
        fp_value = FixedPoint(-19.5, total_bits=16, frac_bits=8)
        actual_bin = list(fp_value.bin_iter())[::-1]
        expected_bin = [1, 1, 1, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(expected_bin, actual_bin)

    def test_neg_0(self):
        fp_value = FixedPoint(0, total_bits=8, frac_bits=2)
        self.assertEqual(0, float(-fp_value))

    def test_neg_minus_5(self):
        fp_value = FixedPoint(-5, total_bits=8, frac_bits=2)
        self.assertEqual(5, float(-fp_value))

    def test_neg_9(self):
        fp_value = FixedPoint(9, total_bits=8, frac_bits=2)
        self.assertEqual(-9, float(-fp_value))

    def test_invert_6(self):
        fp_value = FixedPoint(6, total_bits=8, frac_bits=2)
        self.assertEqual(231, int(~fp_value))

    def test_invert_minus_15(self):
        fp_value = FixedPoint(-15, total_bits=8, frac_bits=2)
        self.assertEqual(59, int(~fp_value))

    def test_abs_minus_5(self):
        fp_value = FixedPoint(-5, total_bits=8, frac_bits=2)
        self.assertEqual(5, abs(fp_value))

    def test_abs_5(self):
        fp_value = FixedPoint(5, total_bits=8, frac_bits=2)
        self.assertEqual(5, abs(fp_value))

    def test_eq(self):
        to_fp = partial(FixedPoint, total_bits=8, frac_bits=4)
        self.assertEqual(to_fp(5.9), to_fp(5.9))

    def test_ne(self):
        to_fp = partial(FixedPoint, total_bits=8, frac_bits=4)
        self.assertNotEqual(to_fp(4.9), to_fp(4.2))

    def test_lt(self):
        to_fp = partial(FixedPoint, total_bits=8, frac_bits=4)
        self.assertLess(to_fp(4.2), to_fp(4.9))

    def test_le(self):
        to_fp = partial(FixedPoint, total_bits=8, frac_bits=4)
        self.assertLessEqual(to_fp(4.2), to_fp(4.9))
        self.assertLessEqual(to_fp(4.2), to_fp(4.2))

    def test_gt(self):
        to_fp = partial(FixedPoint, total_bits=8, frac_bits=4)
        self.assertGreater(to_fp(4.9), to_fp(4.2))

    def test_ge(self):
        to_fp = partial(FixedPoint, total_bits=8, frac_bits=4)
        self.assertGreaterEqual(to_fp(4.9), to_fp(4.2))
        self.assertGreaterEqual(to_fp(4.9), to_fp(4.9))

    def test_add(self):
        fp1 = FixedPoint(-2.5, total_bits=8, frac_bits=4)
        fp2 = FixedPoint(5, total_bits=8, frac_bits=4)
        result = FixedPoint(2.5, total_bits=8, frac_bits=4)
        self.assertEqual(result, fp1 + fp2)

    def test_add_overflow(self):
        fp1 = FixedPoint(4, total_bits=8, frac_bits=4)
        fp2 = FixedPoint(5, total_bits=8, frac_bits=4)
        result = FixedPoint(-7, total_bits=8, frac_bits=4)
        self.assertEqual(result, fp1 + fp2)

    def test_add_incompatible(self):
        fp1 = FixedPoint(3, total_bits=8, frac_bits=4)
        fp2 = FixedPoint(2.5, total_bits=8, frac_bits=2)
        with self.assertRaises(ValueError):
            _ = fp1 + fp2

    def test_sub(self):
        fp1 = FixedPoint(3, total_bits=8, frac_bits=4)
        fp2 = FixedPoint(4.5, total_bits=8, frac_bits=4)
        result = FixedPoint(-1.5, total_bits=8, frac_bits=4)
        self.assertEqual(result, fp1 - fp2)

    def test_sub_overflow(self):
        fp1 = FixedPoint(-4, total_bits=8, frac_bits=4)
        fp2 = FixedPoint(5, total_bits=8, frac_bits=4)
        result = FixedPoint(7, total_bits=8, frac_bits=4)
        self.assertEqual(result, fp1 - fp2)

    def test_sub_incompatible(self):
        fp1 = FixedPoint(3, total_bits=8, frac_bits=4)
        fp2 = FixedPoint(2.5, total_bits=8, frac_bits=2)
        with self.assertRaises(ValueError):
            _ = fp1 - fp2

    def test_and(self):
        fp1 = FixedPoint(5, total_bits=4, frac_bits=0)
        fp2 = FixedPoint(6, total_bits=4, frac_bits=0)
        result = FixedPoint(4, total_bits=4, frac_bits=0)
        self.assertEqual(result, fp1 & fp2)

    def test_or(self):
        fp1 = FixedPoint(5, total_bits=4, frac_bits=0)
        fp2 = FixedPoint(6, total_bits=4, frac_bits=0)
        result = FixedPoint(7, total_bits=4, frac_bits=0)
        self.assertEqual(result, fp1 | fp2)

    def test_xor(self):
        fp1 = FixedPoint(5, total_bits=4, frac_bits=0)
        fp2 = FixedPoint(6, total_bits=4, frac_bits=0)
        result = FixedPoint(3, total_bits=4, frac_bits=0)
        self.assertEqual(result, fp1 ^ fp2)

    def test_to_signed_int_positive_value(self):
        fp = FixedPoint(5, total_bits=8, frac_bits=4)
        self.assertEqual(80, fp.to_signed_int())

    def test_to_signed_int_negative_value(self):
        fp = FixedPoint(-5, total_bits=8, frac_bits=4)
        self.assertEqual(-80, fp.to_signed_int())


class InferTotalAndFracBits(TestCase):
    def test_infer_empty_list(self):
        with self.assertRaises(ValueError):
            _, _ = infer_total_and_frac_bits([])

    def test_infer_mixed_total_bits(self):
        with self.assertRaises(ValueError):
            _, _ = infer_total_and_frac_bits(
                [
                    FixedPoint(0, total_bits=8, frac_bits=4),
                    FixedPoint(0, total_bits=8, frac_bits=4),
                    FixedPoint(0, total_bits=12, frac_bits=4),
                    FixedPoint(0, total_bits=8, frac_bits=4),
                ]
            )

    def test_infer_mixed_frac_bits(self):
        with self.assertRaises(ValueError):
            _, _ = infer_total_and_frac_bits(
                [
                    FixedPoint(0, total_bits=8, frac_bits=4),
                    FixedPoint(0, total_bits=8, frac_bits=4),
                    FixedPoint(0, total_bits=8, frac_bits=5),
                    FixedPoint(0, total_bits=8, frac_bits=4),
                ]
            )

    def test_infer_mixed_total_and_frac_bits(self):
        with self.assertRaises(ValueError):
            _, _ = infer_total_and_frac_bits(
                [
                    FixedPoint(0, total_bits=8, frac_bits=4),
                    FixedPoint(0, total_bits=8, frac_bits=4),
                    FixedPoint(0, total_bits=8, frac_bits=5),
                    FixedPoint(0, total_bits=12, frac_bits=4),
                ]
            )

    def test_infer_valid_list(self):
        total_bits, frac_bits = infer_total_and_frac_bits(
            [
                FixedPoint(0, total_bits=8, frac_bits=4),
                FixedPoint(0, total_bits=8, frac_bits=4),
            ]
        )
        self.assertEqual(8, total_bits)
        self.assertEqual(4, frac_bits)

    def test_infer_multiple_valid_lists(self):
        values = [
            FixedPoint(0, total_bits=8, frac_bits=4),
            FixedPoint(0, total_bits=8, frac_bits=4),
        ]
        total_bits, frac_bits = infer_total_and_frac_bits(values, values, values)
        self.assertEqual(8, total_bits)
        self.assertEqual(4, frac_bits)

    def test_infer_multiple_invalid_lists(self):
        values1 = [
            FixedPoint(0, total_bits=8, frac_bits=4),
            FixedPoint(0, total_bits=8, frac_bits=4),
        ]
        values2 = [
            FixedPoint(0, total_bits=12, frac_bits=4),
            FixedPoint(0, total_bits=8, frac_bits=4),
        ]
        with self.assertRaises(ValueError):
            _, _ = infer_total_and_frac_bits(values1, values2)

    def test_infer_multiple_lists_with_empty_list(self):
        values = [
            FixedPoint(0, total_bits=8, frac_bits=4),
            FixedPoint(0, total_bits=8, frac_bits=4),
        ]
        with self.assertRaises(ValueError):
            _, _ = infer_total_and_frac_bits(values, [])


class FloatValuesToFixedPointTest(TestCase):
    def test_empty_list(self):
        actual = float_values_to_fixed_point([], total_bits=8, frac_bits=4)
        self.assertListEqual([], actual)

    def test_full_list(self):
        actual = float_values_to_fixed_point(
            values=[1, 2, 3],
            total_bits=8,
            frac_bits=4,
        )
        target = [FixedPoint(value, total_bits=8, frac_bits=4) for value in range(1, 4)]
        self.assertListEqual(target, actual)


class IntValuesToFixedPointTest(TestCase):
    def test_empty_list(self):
        actual = float_values_to_fixed_point([], total_bits=8, frac_bits=0)
        self.assertListEqual([], actual)

    def test_full_list(self):
        actual = float_values_to_fixed_point(
            values=[1, 2, 3],
            total_bits=8,
            frac_bits=0,
        )
        target = [FixedPoint(value, total_bits=8, frac_bits=4) for value in range(1, 4)]
        self.assertListEqual(target, actual)


class NumberEncoderTest(TestCase):
    """
    Test Cases:
      - build new encoder from existing encoder ensuring compatibility of enumerations
        Use case scenario: Connecting the outputs of layer h_1 to the inputs of layer h_2, while we can consider the in-
         and output as enumerations, ie. we don't care about the actual numeric values. However to still allow for
         max pooling operations we might want to ensure that the encoding is monotonous.
    """

    def test_binarization_minus_one_is_zero(self):
        encoder = ToLogicEncoder()
        encoder.register_symbol(-1)
        self.assertEqual(0, encoder[-1])

    def test_binarization_minus_1_to_0_and_1to1(self):
        encoder = ToLogicEncoder()
        encoder.register_symbol(-1)
        encoder.register_symbol(1)
        self.assertEqual(1, encoder[1])

    def test_encoder_is_monotonous(self):
        encoder = ToLogicEncoder()
        encoder.register_symbol(1)
        encoder.register_symbol(-1)
        self.assertEqual(1, encoder[1])

    def test_encoder_to_bit_vector(self):
        encoder = ToLogicEncoder()
        encoder.register_symbol(1)
        encoder.register_symbol(-1)
        bits = encoder(-1)
        self.assertEqual("0", bits)

    def test_ternarization_minus1_to_00(self):
        encoder = ToLogicEncoder()
        encoder.register_symbol(-1)
        encoder.register_symbol(0)
        encoder.register_symbol(1)
        test_parameters = (
            (0, -1),
            (1, 0),
            (2, 1),
        )
        for parameter in test_parameters:
            with self.subTest(parameter):
                expected = parameter[0]
                actual = encoder[parameter[1]]
                self.assertEqual(expected, actual)

    def test_registering_numbers_in_batch(self):
        one_by_one = ToLogicEncoder()
        by_batch = ToLogicEncoder()
        batch = [1, 0, -1]
        for number in batch:
            one_by_one.register_symbol(number)
        by_batch.register_symbols(batch)
        self.assertTrue(
            by_batch == one_by_one,
            "expected: {}, actual: {}".format(one_by_one, by_batch),
        )



import unittest
from io import BytesIO

import onnx
import torch

from elasticai.creator.onnx import ModuleWrapper
from elasticai.creator.tags_utils import tag
from elasticai.creator.tests.tensor_test_case import TensorTestCase

ONNX_HEADER = """ir_version: 4
producer_name: "pytorch"
producer_version: "1.11.0"
"""


class OnnxExportTest(TensorTestCase):
    """
    For documentation of the model returned by the onnx load function refer to
    https://github.com/onnx/onnx/blob/master/docs/IR.md

    TODO:
      - get rid of the warnings: The shape inference of elasticai.creator::Wrapper type is missing, so it may result in
        wrong shape inference for the exported graph. Please consider adding it in symbolic function.
    """

    def test_can_export_and_load(self):
        module = torch.nn.Sigmoid()
        module = ModuleWrapper(module)
        expected = self.get_simple_string_representation(
            operation_name=type(module.module).__name__, domain="elasticai.creator"
        )
        with BytesIO() as buffer:
            torch.onnx.export(
                module,
                torch.ones(
                    1,
                ),
                buffer,
            )
            buffer.seek(0)
            model = onnx.load(buffer)

        self.assertEqual(expected, "{}".format(model))

    @classmethod
    def get_simple_string_representation(cls, operation_name, domain) -> str:
        template = (
            ONNX_HEADER
            + """graph {
  node {
    output: "1"
    name: "Wrapper_0"
    op_type: "Wrapper"
"""
        )
        attributes = cls.build_attributes(
            """name: "operation_name"
      s: "{operation_name}"
      type: STRING""".format(
                operation_name=operation_name
            )
        )
        template = (
            template
            + attributes
            + """    domain: "elasticai.creator"
  }
  name: "torch-jit-export"
  output {
    name: "1"
    type {
      tensor_type {
        elem_type: 1
        shape {
          dim {
            dim_param: "Wrapper1_dim_0"
          }
        }
      }
    }
  }
}
opset_import {
  version: 9
}
opset_import {
  domain: "elasticai.creator"
  version: 1
}
"""
        )
        return template

    def test_with_tag_int(self):
        model = torch.nn.Sigmoid()
        input_shape = [3, 3]
        model = ModuleWrapper(tag(model, input_shape=input_shape))
        expected = self.get_string_representation_with_tag(
            operation_name=type(model.module).__name__,
            input_shape="""      ints: 3
      ints: 3
      type: INTS""",
        )

        self.check_stringified_onnx_model(expected_string=expected, model=model)

    def test_with_tag_float(self):
        model = torch.nn.Sigmoid()
        input_shape = [3.5, 3.5]
        model = ModuleWrapper(tag(model, input_shape=input_shape))
        expected = self.get_string_representation_with_tag(
            operation_name="Sigmoid",
            input_shape="""      floats: 3.5
      floats: 3.5
      type: FLOATS""",
        )

        self.check_stringified_onnx_model(expected_string=expected, model=model)

    def test_with_tag_Tensor(self):
        model = torch.nn.Sigmoid()
        input_shape = torch.tensor([1])
        model = ModuleWrapper(tag(model, input_shape=input_shape))
        expected = self.get_string_representation_with_tag(
            operation_name=type(model.module).__name__,
            input_shape="""      t {
        dims: 1
        data_type: 7
        raw_data: "\\001\\000\\000\\000\\000\\000\\000\\000"
      }
      type: TENSOR""",
        )

        self.check_stringified_onnx_model(expected_string=expected, model=model)

    def test_with_tag_String(self):
        model = torch.nn.Sigmoid()
        input_shape = "abc"
        model = ModuleWrapper(tag(model, input_shape=input_shape))
        expected = self.get_string_representation_with_tag(
            operation_name=type(model.module).__name__,
            input_shape="""      s: "abc"
      type: STRING""",
        )

        self.check_stringified_onnx_model(expected_string=expected, model=model)

    def check_stringified_onnx_model(self, expected_string, model):
        with BytesIO() as buffer:
            torch.onnx.export(
                model,
                torch.ones(
                    1,
                ),
                buffer,
            )
            buffer.seek(0)
            model = onnx.load(buffer)
        self.assertEqual(expected_string, "{}".format(model))

    @staticmethod
    def get_string_representation_with_tag(operation_name, input_shape) -> str:
        template = (
            ONNX_HEADER
            + """graph {{
  node {{
    output: "1"
    name: "Wrapper_0"
    op_type: "Wrapper"
    attribute {{
      name: "input_shape"
{input_shape}
    }}
    attribute {{
      name: "operation_name"
      s: "{operation_name}"
      type: STRING
    }}
    domain: "elasticai.creator"
  }}
  name: "torch-jit-export"
  output {{
    name: "1"
    type {{
      tensor_type {{
        elem_type: 1
        shape {{
          dim {{
            dim_param: "Wrapper1_dim_0"
          }}
        }}
      }}
    }}
  }}
}}
opset_import {{
  version: 9
}}
opset_import {{
  domain: "elasticai.creator"
  version: 1
}}
"""
        )
        template = template.format(
            operation_name=operation_name, input_shape=input_shape
        )
        return template

    @staticmethod
    def build_attributes(*args: str) -> str:
        attribute_string = ""
        for attribute in args:
            attribute_string += "    attribute {{\n      {attribute}\n    }}\n".format(
                attribute=attribute
            )
        return attribute_string

    def test_wrapper_transparent(self):
        module = torch.nn.Sigmoid()
        module = ModuleWrapper(module)
        input = torch.rand([2, 2])
        expected = module.module(input)
        actual = module(input)
        self.assertTensorEquals(expected, actual)


if __name__ == "__main__":
    unittest.main()



from importlib import resources
from importlib.abc import Traversable
from pathlib import PurePath

PathType = str | PurePath


def _get_file(package: resources.Package, file_name: str) -> Traversable:
    for resource in resources.files(package).iterdir():
        if resource.name == file_name:
            return resource
    raise FileNotFoundError(f"The file '{file_name}' does not exist.")


def read_text(
    package: resources.Package, file_name: str, encoding: str = "utf-8"
) -> str:
    return _get_file(package, file_name).read_text(encoding)


def read_bytes(package: resources.Package, file_name: str) -> bytes:
    return _get_file(package, file_name).read_bytes()


def copy_file(
    package: resources.Package, file_name: str, destination: PathType
) -> None:
    data = read_bytes(package, file_name)
    with open(destination, "wb") as out_file:
        out_file.write(data)


def get_full_path(package: resources.Package, file_name: str) -> str:
    resource = _get_file(package, file_name)
    with resources.as_file(resource) as file:
        return file.as_posix()



import doctest
from unittest import TestCase

import elasticai.creator.vhdl.templates.utils
from elasticai.creator.vhdl.templates.utils import expand_multiline_template


class ExpandTemplatesTestCase(TestCase):
    @staticmethod
    def get_result_string(template, **kwargs) -> str:
        return "\n".join(expand_multiline_template(template, **kwargs))

    def test_expand_multiline_with_one_item(self):
        # noinspection PyShadowingNames
        def check_single_item(value: str):
            template = "$my_var"
            expected = value
            values = [value]
            actual = self.get_result_string(template, my_var=values)
            self.assertEqual(expected, actual)

        for value in ["value", "some other value"]:
            with self.subTest():
                check_single_item(value)

    def single_value_with_different_key(self):
        template = "$my_key"
        expected = "value"
        actual = self.get_result_string(template, my_key=[expected])
        self.assertEqual(expected, actual)

    def test_expand_two_keys(self):
        template = "$first\n$second"
        expected = "ab\ncd"
        actual = "\n".join(
            list(expand_multiline_template(template, first=("ab",), second=("cd",)))
        )
        self.assertEqual(expected, actual)

    def test_two_values(self):
        template = "$key"
        expected = "a\nb"
        actual = self.get_result_string(template, key=["a", "b"])
        self.assertEqual(expected, actual)

    def test_no_values(self):
        template = "$key"
        expected = ""
        actual = self.get_result_string(template, key=[])
        self.assertEqual(expected, actual)

    def test_key_not_in_template(self):
        template = "something"
        expected = "something"
        actual = self.get_result_string(template, key=["a"])
        self.assertEqual(expected, actual)

    def test_keep_indentation(self):
        template = "\t  $key"
        expected = "\t  a\n\t  b\n\t  c"
        actual = "\n".join(
            list(expand_multiline_template(template, key=["a", "b", "c"]))
        )
        self.assertEqual(expected, actual)

    def test_no_error_if_not_all_keys_are_filled(self):
        template = "$key"
        # noinspection PyBroadException
        try:
            actual = self.get_result_string(template, other_key=[])
            self.assertEqual(template, actual)
        except Exception:
            self.fail()

    def test_two_keys_on_one_line(self):
        template = "\t$first $second"
        values = [0, 1]
        expected = "\t0 $second\n\t1 $second"
        actual = self.get_result_string(template, first=values, second=values)
        self.assertEqual(expected, actual)

    def test_expand_two_keys_by_running_twice(self):
        template = "\t$first $second"
        values = [0, 1]
        expected = "\t0 0\n\t0 1\n\t1 0\n\t1 1"
        actual = self.get_result_string(template, first=values)
        actual = self.get_result_string(actual, first=values, second=values)
        self.assertEqual(expected, actual)


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(elasticai.creator.vhdl.templates.utils))
    return tests



import unittest


class GeneratedVHDLCodeTest(unittest.TestCase):
    @staticmethod
    def get_non_empty_lines(text):
        text = text.splitlines()
        return [x for x in text if len(x) > 0]

    def check_generated_code(self, expected_code, code, check_comments: bool = False):
        expected = self.get_non_empty_lines(expected_code)
        actual = self.get_non_empty_lines(code)
        if check_comments:
            expected = [line.strip() for line in expected if not line.isspace()]

            actual = [line.strip() for line in actual if not line.isspace()]
        else:
            expected = [
                line.strip()
                for line in expected
                if not line.startswith("--") and not line.isspace()
            ]

            actual = [
                line.strip()
                for line in actual
                if not line.startswith("--") and not line.isspace()
            ]
        self.assertEqual(expected, actual)

    expected_code = None



import numpy as np
from torch import Tensor, nn

from elasticai.creator.precomputation.precomputation import (
    get_precomputations_from_direct_children,
    precomputable,
)

model = nn.Sequential(
    precomputable(
        nn.Sigmoid(),
        input_shape=(-1,),
        input_generator=lambda input_shape: Tensor(
            np.linspace(-10, 10, num=20).reshape(input_shape)
        ),
    )
)

for child in model.named_children():
    print(child)

precomputations = get_precomputations_from_direct_children(model)
for precomputation in precomputations:
    precomputation()
    print("x: ", precomputation[0], "\ny: ", precomputation[1])



import math
from collections.abc import Sequence
from itertools import chain
from typing import Any, Iterable, Iterator


class FixedPoint:
    """
    A data type that converts a given number to the corresponding fixed-point representation.
    A fixed-point value is an unsigned integer in two's complement.

    Parameters:
        value (float | int): Value to be represented as fixed-point value.
        total_bits (int): Total number of bits of the fixed-point representation (including number of fractional bits).
        frac_bits (int): Number of bits to represent the fractional part of the number.

    Examples:
        >>> fixed_point_value = FixedPoint(-2.9, total_bits=8, frac_bits=4)
        >>> int(fixed_point_value)
        210
        >>> float(fixed_point_value)
        -2.875
        >>> fixed_point_value.to_signed_int()
        -46
        >>> fixed_point_value.to_bin()
        '11010010'
        >>> fixed_point_value.to_hex()
        'd2'
    """

    __slots__ = ["_value", "_frac_bits", "_total_bits"]

    def __init__(
        self,
        value: float | int,
        total_bits: int,
        frac_bits: int,
    ) -> None:
        self._value = float(value)
        self._total_bits = total_bits
        self._frac_bits = frac_bits
        FixedPoint._assert_range(self._value, self._total_bits, self._frac_bits)

    def __int__(self) -> int:
        fp_int = int(self._value * (1 << self._frac_bits))
        if fp_int < 0:
            fp_int = FixedPoint._calculate_two_complement(fp_int, self._total_bits)
        return fp_int

    def __float__(self) -> float:
        return FixedPoint._calculate_float_from_fixed_point(
            int(self), self._total_bits, self._frac_bits
        )

    def __eq__(self, other: Any) -> bool:
        return float(self) == float(other)

    def __ne__(self, other: Any) -> bool:
        return float(self) != float(other)

    def __lt__(self, other: Any) -> bool:
        return float(self) < float(other)

    def __le__(self, other: Any) -> bool:
        return float(self) <= float(other)

    def __gt__(self, other: Any) -> bool:
        return float(self) > float(other)

    def __ge__(self, other: Any) -> bool:
        return float(self) >= float(other)

    def __add__(self, other: "FixedPoint") -> "FixedPoint":
        FixedPoint._assert_is_compatible(self, other)
        return self._identical_fixed_point_from_int(
            FixedPoint._discard_leading_bits(
                int(self) + int(other), num_bits=self._total_bits
            )
        )

    def __sub__(self, other: "FixedPoint") -> "FixedPoint":
        return self + (-other)

    def __and__(self, other: "FixedPoint") -> "FixedPoint":
        FixedPoint._assert_is_compatible(self, other)
        return self._identical_fixed_point_from_int(int(self) & int(other))

    def __or__(self, other: "FixedPoint") -> "FixedPoint":
        FixedPoint._assert_is_compatible(self, other)
        return self._identical_fixed_point_from_int(int(self) | int(other))

    def __xor__(self, other: "FixedPoint") -> "FixedPoint":
        FixedPoint._assert_is_compatible(self, other)
        return self._identical_fixed_point_from_int(int(self) ^ int(other))

    def __invert__(self) -> "FixedPoint":
        return self._identical_fixed_point_from_int(
            FixedPoint._invert_int(int(self), num_bits=self._total_bits)
        )

    def __neg__(self) -> "FixedPoint":
        return self._identical_fixed_point(-self._value)

    def __abs__(self) -> "FixedPoint":
        return self._identical_fixed_point(abs(self._value))

    def __str__(self) -> str:
        return str(int(self))

    def __repr__(self) -> str:
        return f"FixedPoint(value={self._value}, total_bits={self._total_bits}, frac_bits={self._frac_bits})"

    def _identical_fixed_point(self, value: float | int) -> "FixedPoint":
        return FixedPoint(
            value=value, total_bits=self._total_bits, frac_bits=self._frac_bits
        )

    def _identical_fixed_point_from_int(self, value: int) -> "FixedPoint":
        return FixedPoint.from_int(
            value=value, total_bits=self._total_bits, frac_bits=self._frac_bits
        )

    @staticmethod
    def _assert_range(value: float | int, total_bits: int, frac_bits: int) -> None:
        max_value = 2 ** (total_bits - frac_bits - 1)
        min_value = max_value * (-1)

        if not min_value <= value < max_value:
            raise ValueError(
                (
                    f"Value {value} cannot represented as a fixed point value with {total_bits} total bits "
                    f"and {frac_bits} fraction bits (value range: [{min_value}, {max_value}))."
                )
            )

    @staticmethod
    def _assert_is_compatible(fp1: "FixedPoint", fp2: "FixedPoint") -> None:
        if not (fp1.total_bits == fp2.total_bits and fp1.frac_bits == fp2.frac_bits):
            raise ValueError(
                (
                    f"FixedPoint objects not compatible (total_bits: {fp1.total_bits} != {fp2.total_bits}); "
                    f"frac_bits: {fp1.frac_bits} != {fp2.frac_bits})."
                )
            )

    @staticmethod
    def _invert_int(value: int, num_bits: int) -> int:
        return value ^ int("1" * num_bits, 2)

    @staticmethod
    def _discard_leading_bits(value: int, num_bits: int) -> int:
        return value & int("1" * num_bits, 2)

    @staticmethod
    def _calculate_two_complement(value: int, num_bits: int) -> int:
        return FixedPoint._invert_int(abs(value), num_bits) + 1

    @staticmethod
    def _calculate_float_from_fixed_point(
        value: int, total_bits: int, frac_bits: int
    ) -> float:
        if value > 2**total_bits - 1:
            raise ValueError(
                f"Value {value} cannot interpreted as a fixed point with {total_bits} total bits."
            )
        is_negative = value & (1 << total_bits - 1) > 0
        if is_negative:
            value = FixedPoint._calculate_two_complement(value, total_bits)
            value *= -1
        return value / (1 << frac_bits)

    @staticmethod
    def from_int(
        value: int,
        total_bits: int,
        frac_bits: int,
    ) -> "FixedPoint":
        return FixedPoint(
            value=FixedPoint._calculate_float_from_fixed_point(
                value, total_bits=total_bits, frac_bits=frac_bits
            ),
            total_bits=total_bits,
            frac_bits=frac_bits,
        )

    @property
    def total_bits(self) -> int:
        return self._total_bits

    @property
    def frac_bits(self) -> int:
        return self._frac_bits

    def bin_iter(self) -> Iterator[int]:
        return ((int(self) >> i) & 1 for i in range(self._total_bits))

    def to_signed_int(self) -> int:
        return int(abs(self)) * (-1 if self < 0 else 1)

    def to_bin(self) -> str:
        return f"{int(self):0{self._total_bits}b}"

    def to_hex(self) -> str:
        return f"{int(self):0{math.ceil(self._total_bits / 4)}x}"


def infer_total_and_frac_bits(*values: Sequence[FixedPoint]) -> tuple[int, int]:
    if sum(len(value_list) == 0 for value_list in values) > 0:
        raise ValueError("Cannot infer total bits and frac bits from an empty list.")
    total_bits, frac_bits = values[0][0].total_bits, values[0][0].frac_bits
    for value in chain(*values):
        if value.total_bits != total_bits or value.frac_bits != frac_bits:
            raise ValueError(
                "Cannot infer total bits and frac bits from a list with mixed total bits or frac bits."
            )
    return total_bits, frac_bits


def float_values_to_fixed_point(
    values: list[float | int], total_bits: int, frac_bits: int
) -> list[FixedPoint]:
    return list(map(lambda x: FixedPoint(x, total_bits, frac_bits), values))


def int_values_to_fixed_point(
    values: list[int], total_bits: int, frac_bits: int
) -> list[FixedPoint]:
    return list(map(lambda x: FixedPoint.from_int(x, total_bits, frac_bits), values))


class ToLogicEncoder:
    """
    Throughout our implementations we have to deal with two different levels of representations for numbers:
    During training we typically need to apply mathematical operations and we do not care too much about how our numbers are encoded.
    E.g. in a scenario where we want to use two bit on hardware to represent our numbers, in our machine learning framework we
    might decide it is beneficial to use the numbers -3 and 4 for some reason. However, especially in the context of precomputed
    results, the hardware implementation does not need to know the numeric values, but instead just needs to be able to keep a
    consistent and correct mapping. The NumericToLogicEncoder takes care of performing the translations from numeric representation
    to the bit vector used in the hardware implementation. We encode bit vectors just as unsigned integers.
    """

    def __init__(self):
        self._symbols = set()
        self._mapping = dict()

    def _update_mapping(self) -> None:
        sorted_numerics = sorted(self._symbols)
        mapping = {value: index for index, value in enumerate(sorted_numerics)}
        self._mapping.update(mapping)

    def register_symbol(self, numeric_representation: int) -> None:
        self._symbols.add(numeric_representation)
        self._update_mapping()

    def register_symbols(self, symbols: Iterable[int]) -> None:
        for symbol in symbols:
            self._symbols.add(symbol)
        self._update_mapping()

    @staticmethod
    def _int_to_bin(number: int, num_bits: int) -> str:
        return f"{number:0{num_bits}b}"

    @property
    def bit_width(self) -> int:
        return math.floor(math.log2(len(self._symbols)))

    def __len__(self):
        return len(self._symbols)

    def __eq__(self, other: "ToLogicEncoder") -> bool:
        return self._symbols == other._symbols and self._mapping == other._mapping

    def __iter__(self) -> Iterator[tuple[int, int]]:
        for symbol, encoded_symbol in self._mapping.values():
            yield symbol, encoded_symbol

    def __getitem__(self, item: int) -> int:
        return self._mapping[item]

    def __call__(self, number: int) -> str:
        if number not in self._symbols:
            raise ValueError
        return self._int_to_bin(self._mapping[number], self.bit_width)



"""Interfaces/Adapters to the used machine learning framework"""

from typing import Iterable, Protocol, runtime_checkable

from numpy.typing import ArrayLike

Index = int | slice | tuple[int | slice]


class Tensor(Protocol):
    def __getitem__(self, index: Index) -> "Tensor":
        ...

    def detach(self) -> "Tensor":
        ...

    def numpy(self) -> ArrayLike:
        ...


@runtime_checkable
class Module(Protocol):
    @property
    def training(self) -> bool:
        ...

    def extra_repr(self) -> str:
        ...

    def named_children(self) -> Iterable[tuple[str, "Module"]]:
        ...

    def __call__(self, x: Tensor, *args, **kwargs) -> Tensor:
        ...



import unittest
from abc import abstractmethod
from typing import Protocol, runtime_checkable
from unittest import TestCase

from elasticai.creator.precomputation.metaprogramming import (
    add_instance_attribute,
    add_method,
)


class TestWrappingClasses(TestCase):
    @unittest.SkipTest
    def test_method(self):
        def my_method(self) -> list[int]:
            return [1]

        @add_method(method_name="my_method", method=my_method)
        class A:
            ...

        @runtime_checkable
        class P(Protocol):
            @abstractmethod
            def my_method(self) -> list[int]:
                ...

        # below assignment should be correct for static type analysis
        a: P = A()

        self.assertTrue(isinstance(a, P))

    def test_attribute(self):
        @add_instance_attribute(attribute_name="my_attribute", attribute=4)
        class A:
            ...

        self.assertTrue(not hasattr(A, "my_attribute"))
        self.assertEqual(4, A().my_attribute)



import unittest

import torch

from elasticai.creator.qat.masks import fixed_offset_mask_4d, random_mask_4d


class test_masks(unittest.TestCase):
    def test_fixedOffsetMask4D_channel_offset(self):
        mask = fixed_offset_mask_4d(
            out_channels=4,
            kernel_size=(1, 1),
            in_channels=2,
            groups=1,
            axis_width=1,
            offset_axis=1,
        )
        expected = torch.zeros((4, 2, 1, 1))
        expected[0, 0, 0, 0] = 1
        expected[1, 1, 0, 0] = 1
        expected[2, 0, 0, 0] = 1
        expected[3, 1, 0, 0] = 1
        self.assertTrue(torch.all(torch.eq(mask, expected)))

    def test_fixedOffsetMask4D_channel_offset_width2(self):
        mask = fixed_offset_mask_4d(
            out_channels=2,
            kernel_size=(1, 1),
            in_channels=4,
            groups=1,
            axis_width=2,
            offset_axis=1,
        )
        expected = torch.zeros((2, 4, 1, 1))
        expected[0, 0, 0, 0] = 1
        expected[0, 1, 0, 0] = 1
        expected[1, 2, 0, 0] = 1
        expected[1, 3, 0, 0] = 1
        self.assertTrue(torch.all(torch.eq(mask, expected)))

    def test_fixedOffsetMask4D_axis_offset(self):
        mask = fixed_offset_mask_4d(
            out_channels=4,
            kernel_size=(2, 1),
            in_channels=2,
            groups=1,
            axis_width=1,
            offset_axis=2,
        )
        expected = torch.zeros((4, 2, 2, 1))
        expected[0, 0, 0, 0] = 1
        expected[0, 1, 0, 0] = 1
        expected[1, 0, 1, 0] = 1
        expected[1, 1, 1, 0] = 1
        expected[2, 0, 0, 0] = 1
        expected[2, 1, 0, 0] = 1
        expected[3, 0, 1, 0] = 1
        expected[3, 1, 1, 0] = 1
        self.assertTrue(torch.all(torch.eq(mask, expected)))

    def test_randomMask4D(self):
        mask = random_mask_4d(
            out_channels=4,
            kernel_size=(2, 2),
            in_channels=2,
            groups=1,
            params_per_channel=2,
        )
        for channel in range(4):
            self.assertEqual(torch.count_nonzero(mask[channel]), 2)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['elasticai',
 'elasticai.creator',
 'elasticai.creator.examples',
 'elasticai.creator.examples.generate_vhdl_files',
 'elasticai.creator.integrationTests',
 'elasticai.creator.integrationTests.vhdl',
 'elasticai.creator.mlframework',
 'elasticai.creator.precomputation',
 'elasticai.creator.qat',
 'elasticai.creator.tests',
 'elasticai.creator.tests.precomputation',
 'elasticai.creator.tests.qat',
 'elasticai.creator.tests.vhdl',
 'elasticai.creator.vhdl',
 'elasticai.creator.vhdl.precomputed_convs',
 'elasticai.creator.vhdl.templates',
 'elasticai.creator.vhdl.templates.precomputed_convs',
 'elasticai.creator.vhdl.vhdl_formatter']

package_data = \
{'': ['*']}

install_requires = \
['MonkeyType>=21.5.0,<22.0.0',
 'bitarray>=2.3.5,<3.0.0',
 'matplotlib>=3.5.2,<4.0.0',
 'torch>=1.11.0,<2.0.0',
 'vsg>=3.9.0,<4.0.0']

extras_require = \
{':extra == "onnx"': ['numpy>=1.23.0,<2.0.0'],
 'examples': ['torchvision>=0.12.0,<0.13.0'],
 'onnx': ['protobuf>=3.16,<3.17', 'onnx>=1.12,<2.0'],
 'systemtests': ['torchvision>=0.12.0,<0.13.0']}

setup_kwargs = {
    'name': 'elasticai.creator',
    'version': '0.10.1',
    'description': 'Design, train and compile neural networks optimized specifically for FPGAs.',
    'long_description': '# ElasticAi.creator\n\nDesign, train and compile neural networks optimized specifically for FPGAs.\nObtaining a final model is typically a three stage process.\n* design and train it using the layers provided in the `elasticai.creator.qat` package.\n* translate the model to a target representation, e.g. VHDL\n* compile the intermediate representation with a third party tool, e.g. Xilinx Vivado (TM)\n\nThis version currently only supports parts of VHDL as target representations.\n\nThe project is part of the elastic ai ecosystem developed by the Embedded Systems Department of the University Duisburg-Essen. For more details checkout the slides at [researchgate](https://www.researchgate.net/publication/356372207_In-Situ_Artificial_Intelligence_for_Self-_Devices_The_Elastic_AI_Ecosystem_Tutorial).\n\n\n## Table of contents\n\n- [Users Guide](#users-guide)\n  - [Install](#install)\n- [Developers Guide](#developers-guide)\n  - [Install Dev Dependencies](#install-dev-dependencies)\n\n\n## Users Guide\n\n#### Install\nYou can install the ElasticAI.creator as a dependency using pip:\n```bash\npython3 -m pip install "elasticai.creator"\n```\n\n\n## Structure of the Project\n\nThe structure of the project is as follows.\nThe [creator](elasticai/creator) folder includes all main concepts of our project, especially the qtorch implementation which is our implementation of quantized PyTorch layer.\nIt also includes the supported target representations, like the subfolder [vhdl](elasticai/creator/vhdl) is for the translation to vhdl.\nAdditionally, we have folders for [unit tests](elasticai/creator/tests) and [integration tests](elasticai/creator/integrationTests).\n\n\n## General Limitations\n\nBy now we only support Sequential models for our translations.\n\n## Developers Guide\n### Install Dev Dependencies\n- [poetry](https://python-poetry.org/)\n- recommended:\n  - [pre-commit](https://pre-commit.com/)\n  - [commitlint](https://github.com/conventional-changelog/commitlint) to help following our [conventional commit](https://www.conventionalcommits.org/en/v1.0.0-beta.2/#summary) guidelines\npoetry can be installed in the following way:\n```bash\npip install poetry\npip install pre-commit\npoetry install -D\npre-commit install\nnpm install --save-dev @commitlint/{config-conventional,cli}\nsudo apt install ghdl\n```\n\n### Commit Message Scopes\n\n- **qat**: quantization-aware-training\n  - Examples: `QConv1D`, `QLSTM`, autograd functions, etc.\n- **readme**\n- **precomputation**: entities that deal with the precomputation of ML components\n  - Examples: the `precomputable` decorator or the `IOTable` class\n- **vhdl**: vhdl code generation\n  - Examples: `vhdl.TruthTable`, `vhdl.LSTMCell`\n- **gh-workflow**\n- **pyproject**: changes to the `pyproject.toml` file will typically either update run or dev dependencies\n- **typing**: changing type annotations and changes to code to allow consistent type annotations\n- **pre-commit**\n\n### Adding new translation targets\nNew translation targets should be located in their own folder, e.g. vhdl for translating from any language to vhdl.\nWorkflow for adding a new translation:\n1. Obtain a structure, such as a list in a sequential case, which will describe the connection between every component.\n2. Identify and label relevant structures, in the base cases it can be simply separate layers.\n3. Map each structure to its function which will convert it.\n4. Do such conversions.\n5. Recreate connections based on 1.\n\nEach sub-step should be separable and it helps for testing if common functions are wrapped around an adapter.\n\n### Syntax Checking\n\n[GHDL](https://ghdl.github.io/ghdl/) supports a [syntax checking](https://umarcor.github.io/ghdl/using/InvokingGHDL.html#check-syntax-s) which checks the syntax of a vhdl file without generating code.\nThe command is as follows:\n```\nghdl -s path/to/vhdl/file\n```\nSo, for example for checking the sigmoid source vhdl files in our project we can run:\n```\nghdl -s elasticai/creator/vhdl/source/sigmoid.vhd\n```\nFor checking all vhdl files together in our project we can just run:\n```\nghdl -s elasticai/creator/**/*.vhd\n```\n\n### Tests\n\nOur implementation is fully tested with unit, integration and system tests.\nPlease refer to the system tests as examples of how to use the Elastic Ai Creator Translator.\nYou can run one explicit test with the following statement:\n\n```python -m unittest discover -p "test_*.py" elasticai/creator/path/to/test.py```\n\nIf you want to run all tests, give the path to the tests:\n\n```python -m unittest discover -p "test_*.py" elasticai/creator/path/to/testfolder```\n\nYou can also run all tests together:\n\n```python -m unittest discover -p "test_*.py" elasticai/creator/translator/path/to/language/```\n\nIf you want to add more tests please refer to the Test Guidelines in the following.\n\n### Test style Guidelines\n\n#### File IO\nIn general try to avoid interaction with the filesystem. In most cases instead of writing to or reading from a file you can use a StringIO object or a StringReader.\nIf you absolutely have to create files, be sure to use pythons [tempfile](https://docs.python.org/3.9/library/tempfile.html) module and cleanup after the tests.\n\n\n#### Diectory structure and file names\nFiles containing tests for a python module should be located in a test directory for the sake of separation of concerns.\nEach file in the test directory should contain tests for one and only one class/function defined in the module.\nFiles containing tests should be named according to the rubric\n`test_ClassName.py`.\nNext, if needed for more specific tests define a class. Then subclass it, in this class define a setUp method (and possibly tearDown) to create the global environment.\nIt avoids introducing the category of bugs associated with copying and pasting code for reuse.\nThis class should be named similarly to the file name.\nThere\'s a category of bugs that appear if  the initialization parameters defined at the top of the test file are directly used: some tests require the initialization parameters to be changed slightly.\nIts possible to define a parameter and have it change in memory as a result of a test.\nSubsequent tests will therefore throw errors.\nEach class contains methods that implement a test.\nThese methods are named according to the rubric\n`test_name_condition`\n\n#### Unit tests\nIn those tests each functionality of each function in the module is tested, it is the entry point  when adding new functions.\nIt assures that the function behaves correctly independently of others.\nEach test has to be fast, so use of heavier libraries is discouraged.\nThe input used is the minimal one needed to obtain a reproducible output.\nDependencies should be replaced with mocks as needed.\n\n#### Integration Tests\nHere the functions\' behaviour with other modules is tested.\nIn this repository each integration function is in the correspondent folder.\nThen the integration with a single class of the target, or the minimum amount of classes for a functionality, is tested in each separated file.\n\n#### System tests\nThose tests will use every component of the system, comprising multiple classes.\nThose tests include expected use cases and unexpected or stress tests.\n\n#### Adding new functionalities and tests required\nWhen adding new functions to an existing module, add unit tests in the correspondent file in the same order of the module, if a new module is created a new file should be created.\nWhen a bug is solved created the respective regression test to ensure that it will not return.\nProceed similarly with integration tests.\nCreating a new file if a functionality completely different from the others is created e.g. support for a new layer.\nSystem tests are added if support for a new library is added.\n\n#### Updating tests\nIf new functionalities are changed or removed the tests are expected to reflect that, generally the ordering is unit tests -> integration tests-> system tests.\nAlso, unit tests that change the dependencies should be checked, since this system is fairly small the internal dependencies are not always mocked.\n\nreferences: https://jrsmith3.github.io/python-testing-style-guidelines.html\n',
    'author': 'Department Embedded Systems - University Duisburg Essen',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/es-ude/elastic-ai.creator',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import os
from argparse import ArgumentParser
from functools import partial

from elasticai.creator.vhdl.number_representations import float_values_to_fixed_point
from elasticai.creator.vhdl.precomputed_scalar_function import (
    PrecomputedScalarTestBench,
)
from elasticai.creator.vhdl.vhdl_formatter.vhdl_formatter import format_vhdl


def main() -> None:
    arg_parser = ArgumentParser()
    arg_parser.add_argument(
        "--path",
        help="path to folder for generated vhd files",
        required=True,
    )
    args = arg_parser.parse_args()
    if not os.path.isdir(args.path):
        os.mkdir(args.path)
    component_name = "sigmoid_tb"
    file_path = os.path.join(args.path, f"{component_name}.vhd")

    to_fp = partial(float_values_to_fixed_point, total_bits=16, frac_bits=2)

    with open(file_path, "w") as writer:
        sigmoid = PrecomputedScalarTestBench(
            component_name=component_name,
            x_list_for_testing=to_fp([-1281, -1000, -500]),
            y_list_for_testing=to_fp([0, 4, 28]),
        )
        sigmoid_code = sigmoid()
        for line in sigmoid_code:
            writer.write(line + "\n")

    format_vhdl(file_path=file_path)


if __name__ == "__main__":
    main()



from typing import Any, Callable

import torch
from torch.nn import Module

from elasticai.creator.tags_utils import get_tags


def _preprocess_tags(tags: dict):
    processed = {}
    unprocessed = []
    for key, value in tags.items():
        to_check = value
        if isinstance(value, list):
            to_check = value[0]
        if isinstance(to_check, int):
            processed[key + "_i"] = value
        elif isinstance(to_check, float):
            processed[key + "_f"] = value
        elif isinstance(to_check, str):
            processed[key + "_s"] = value
        elif isinstance(to_check, torch.Tensor):
            processed[key + "_t"] = value
        else:
            unprocessed.append(value)
    return processed, unprocessed


class AutogradWrapper(torch.autograd.Function):
    """
    Adds symbolic function so the tags can be exported. Is transparent

    """

    @staticmethod
    def jvp(ctx: Any, *grad_inputs: Any) -> Any:
        raise NotImplementedError

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        raise NotImplementedError

    @staticmethod
    def forward(ctx: Any, input: Any, callable: Callable) -> Any:
        return callable(input)

    @staticmethod
    def symbolic(g, x, wrapped: Module):
        tags = get_tags(wrapped)
        kwargs, args = _preprocess_tags(tags)
        ret = g.op(
            "elasticai.creator::Wrapper",
            *args,
            operation_name_s=type(wrapped).__name__,
            **kwargs
        )
        return ret


class ModuleWrapper(Module):
    """
    Wraps the module so that it applies an autograd

    """

    def __init__(
        self,
        module,
        autograd_fn=AutogradWrapper,
    ):
        super().__init__()
        self.autograd_fn = autograd_fn
        self.module = module

    def forward(self, x):
        return self.autograd_fn.apply(x, self.module)



import unittest

import numpy as np

from elasticai.creator.precomputation.io_table import IOTable


class IOTableBuilderTestCase(unittest.TestCase):
    def test_group_tables_basic(self):
        inputs = np.asarray([[1, 2], [3, 4]])
        outputs = np.asarray([[4, 5], [5, 4]])
        table = IOTable(inputs, outputs)
        result = list(table.grouped(groups=1))
        self.assertTrue(
            np.all(inputs == result[0].inputs),
            np.all(outputs == result[0].outputs),
        )

    def test_group_tables_depthwise(self):
        inputs = np.asarray(
            [
                [
                    1,
                    2,
                ],
                [3, 4],
            ]
        )
        outputs = np.asarray([[1, 2], [6, 5]])
        table = IOTable(inputs, outputs)
        result = table.grouped(groups=2)
        expected = [[], []]
        expected[0] = (np.asarray([[1], [3]]), np.asarray([[1], [6]]))
        expected[1] = (np.asarray([[2], [4]]), np.asarray([[2], [5]]))
        for expected_pair, result_pair in zip(expected, result):
            self.assertTrue(np.all(expected_pair[0] == result_pair.inputs))
            self.assertTrue(np.all(expected_pair[1] == result_pair.outputs))

    def test_group_tables_depthwise_longer_input(self):
        inputs = np.asarray([[1, 2, 3, 4], [3, 4, 5, 6]])
        outputs = np.asarray([[1, 2], [6, 5]])
        table = IOTable(inputs, outputs)
        result = table.grouped(groups=2)
        expected = [[], []]
        expected[0] = (np.asarray([[1, 2], [3, 4]]), np.asarray([[1], [6]]))
        expected[1] = (np.asarray([[3, 4], [5, 6]]), np.asarray([[2], [5]]))
        for expected_pair, result_pair in zip(expected, result):
            self.assertTrue(np.all(expected_pair[0] == result_pair.inputs))
            self.assertTrue(np.all(expected_pair[1] == result_pair.outputs))

    def test_create_io_dict_basic(self) -> None:
        inputs = np.array([[1], [2], [3]])
        outputs = np.array([[2], [3], [4]])
        Iotable = IOTable(inputs, outputs)
        dict = Iotable.get_table_as_dict()
        self.assertEqual(dict, {(1,): (2,), (2,): (3,), (3,): (4,)})

    def test_create_io_dict_depthwise(self) -> None:
        io_list = [
            IOTable(np.asarray([[1, 2], [3, 4]]), np.asarray([[1], [6]])),
            IOTable(np.asarray([[3, 4], [5, 6]]), np.asarray([[2], [5]])),
        ]
        dict_list = io_list[0].get_table_as_dict(), io_list[1].get_table_as_dict()
        self.assertEqual(dict_list[0], {(1, 2): (1,), (3, 4): (6,)})
        self.assertEqual(dict_list[1], {(3, 4): (2,), (5, 6): (5,)})



from types import SimpleNamespace

import torch
from torch import Tensor

from elasticai.creator.qat.functional import binarize
from elasticai.creator.tests.tensor_test_case import TensorTestCase


class BinarizeFunctionTest(TensorTestCase):
    def test_ForwardRaisesErrorOnMissingInput(self):
        def save_for_backward_dummy(_):
            pass

        dummy_context = SimpleNamespace(save_for_backward=save_for_backward_dummy)
        try:
            binarize.forward(dummy_context)
        except TypeError:
            return

        self.fail()

    def test_Yields1For0(self):
        self.assertTensorEquals(
            expected=Tensor([1.0]), actual=binarize.apply(Tensor([0.0]))
        )

    def test_Yields1For2Point4(self):
        self.assertTensorEquals(
            expected=Tensor([1.0]), actual=binarize.apply(Tensor([2.4]))
        )

    def test_YieldMinus1ForNegativeInput(self):
        self.assertTensorEquals(
            expected=Tensor([-1.0]), actual=binarize.apply(Tensor([-2.8]))
        )

    def check_gradient(self, expected_grad, x):
        x = torch.tensor([x], requires_grad=True)
        y = binarize.apply(x)
        y.backward()
        self.assertTensorEquals(torch.tensor([expected_grad]), x.grad)

    def test_gradient_is_0_for_input_greater_1(self):
        self.check_gradient(expected_grad=0.0, x=1.1)

    def test_gradient_is_0_for_input_less_minus_1(self):
        self.check_gradient(expected_grad=0.0, x=-1.1)

    def test_gradient_is_1_for_input_between_minus_1_and_1(self):
        self.check_gradient(expected_grad=1.0, x=-0.8)



from typing import Any, Protocol, TypeVar, runtime_checkable

from elasticai.creator.mlframework import Module

T = TypeVar("T")


@runtime_checkable
class Tagged(Protocol):
    @property
    def elasticai_tags(self) -> dict:
        ...

    @elasticai_tags.setter
    def elasticai_tags(self, tags: dict) -> None:
        ...


T_co = TypeVar("T_co", bound=Module, covariant=True)


@runtime_checkable
class TaggedModule(Tagged, Module, Protocol):
    ...


def tag(module: TaggedModule, **new_tags: Any) -> TaggedModule:
    """Add tags to any object wrapping it in a TagWrapper if necessary

    new_tags will override possibly existing tags
    """
    old_tags = get_tags(module)
    tags = old_tags | new_tags
    module.elasticai_tags = tags
    return module


def get_tags(module: Any) -> dict:
    if isinstance(module, Tagged):
        return module.elasticai_tags
    return {}


def has_tag(module: Any, tag: str) -> bool:
    if isinstance(module, Tagged):
        return tag in module.elasticai_tags
    return False



import pathlib
import re

import torch
from torch import jit
from torch.nn.functional import hardtanh

from elasticai.creator.qat.layers import Binarize
from elasticai.creator.tests.tensor_test_case import TensorTestCase


class JitTests(TensorTestCase):
    def test_script_a_sign_function(self):
        @jit.script
        def function(x):
            return x.sign()

        file_location = pathlib.Path(__file__).resolve()
        expected = f"""graph(%x.1 : Tensor):
  %2 : Tensor = aten::sign(%x.1)
  return (%2)
"""
        actual = f"{function.graph}"
        actual_with_comment_removed = re.sub(r" #(.*)\n", "\n", actual)
        self.assertEqual(expected, actual_with_comment_removed)

    def test_grad_of_clamp_is_zero_if_out_of_range(self):
        x = torch.tensor(-5.0, requires_grad=True)
        y = torch.clamp(x, -1.0, 1.0)
        y.backward()
        expected = torch.tensor(0.0)
        self.assertTensorEquals(expected, x.grad)

    def test_grad_of_clamp_is_one_if_in_range(self):
        x = torch.tensor(-0.5, requires_grad=True)
        y = torch.nn.functional.hardtanh(x)
        y.backward()
        expected = torch.tensor(1.0)
        self.assertTensorEquals(expected, x.grad)

    def test_grad_of_bin_ste(self):
        x = torch.tensor(0.3, requires_grad=True)
        y = Binarize()(x)
        y.backward()
        expected = torch.tensor(1.0)
        self.assertTensorEquals(expected, x.grad)

    def test_grad_of_bin_ste_is_zero(self):
        x = torch.tensor([1.2], requires_grad=True)
        y = Binarize()(x)
        y.backward()
        expected = torch.tensor([0.0])
        self.assertTensorEquals(expected, x.grad)

    def test_value_of_bin_ste(self):
        x = torch.tensor([0.3])
        y = Binarize()(x)
        expected = torch.tensor([1.0])
        self.assertTensorEquals(expected, y)

    def test_negative_value_of_bin_ste(self):
        x = torch.tensor([-0.3])
        y = Binarize()(x)
        expected = torch.tensor([-1.0])
        self.assertTensorEquals(expected, y)

    def test_script_heaviside_based_ste(self):
        function = jit.trace(Binarize(), torch.tensor([0.0]))
        actual = function(torch.tensor([-1.2, -0.4, 8.9]))
        expected = torch.tensor([-1.0, -1.0, 1.0])
        self.assertTensorEquals(expected, actual)



import unittest
from functools import partial

from elasticai.creator.vhdl.number_representations import float_values_to_fixed_point
from elasticai.creator.vhdl.precomputed_scalar_function import (
    precomputed_scalar_function_process,
)


class PrecomputedScalarFunctionProcessTest(unittest.TestCase):
    def setUp(self) -> None:
        self.fp_list = partial(float_values_to_fixed_point, total_bits=16, frac_bits=8)

    def test_empty_x_list_y_list_one_element(self) -> None:
        self.assertEqual(
            list(precomputed_scalar_function_process(x=[], y=self.fp_list([1]))()),
            ['y <= "0000000100000000";'],
        )

    def test_empty_x_list_y_list_too_many_elements(self) -> None:
        with self.assertRaises(ValueError):
            list(precomputed_scalar_function_process(x=[], y=self.fp_list([1, 2]))())

    def test_empty_y_list(self) -> None:
        with self.assertRaises(ValueError):
            list(precomputed_scalar_function_process(x=[], y=[])())

    def test_x_list_lengths_not_suitable_for_y_list_lengths(self) -> None:
        with self.assertRaises(ValueError):
            list(
                precomputed_scalar_function_process(
                    x=self.fp_list([1, 2]), y=self.fp_list([1])
                )()
            )

    def test_x_list_with_only_one_element(self) -> None:
        expected_code = [
            "if int_x<256 then",
            '\ty <= "0000000100000000"; -- 256',
            "else",
            '\ty <= "0000001000000000"; -- 512',
            "end if;",
        ]
        self.assertEqual(
            expected_code,
            list(
                precomputed_scalar_function_process(
                    x=self.fp_list([1]), y=self.fp_list([1, 2])
                )()
            ),
        )

    def test_unsorted_x_list(self) -> None:
        expected_code = [
            "if int_x<256 then",
            '\ty <= "0000000100000000"; -- 256',
            "elsif int_x<512 then",
            '\ty <= "0000001000000000"; -- 512',
            "elsif int_x<768 then",
            '\ty <= "0000001100000000"; -- 768',
            "else",
            '\ty <= "0000010000000000"; -- 1024',
            "end if;",
        ]
        self.assertEqual(
            expected_code,
            list(
                precomputed_scalar_function_process(
                    x=self.fp_list([3, 1, 2]), y=self.fp_list([1, 2, 3, 4])
                )()
            ),
        )



from functools import partial

from elasticai.creator.tests.vhdl.vhdl_file_testcase import GeneratedVHDLCodeTest
from elasticai.creator.vhdl.number_representations import float_values_to_fixed_point
from elasticai.creator.vhdl.precomputed_scalar_function import (
    PrecomputedScalarTestBench,
)


class TanhTestBenchTest(GeneratedVHDLCodeTest):
    def test_compare_files(self) -> None:
        expected_code = """
        library ieee;
        use ieee.std_logic_1164.all;
        use ieee.numeric_std.all;
        use ieee.math_real.all;

        entity tanh_tb is
            generic (
                DATA_WIDTH : integer := 16;
                FRAC_WIDTH : integer := 0
            );
            port (
                clk : out std_logic
            );
        end entity tanh_tb;

        architecture rtl of tanh_tb is

            signal clk_period : time := 1 ns;
            signal test_input : signed(16-1 downto 0):=(others=>'0');
            signal test_output : signed(16-1 downto 0);

            component tanh is
                generic (
                    DATA_WIDTH : integer := 16;
                    FRAC_WIDTH : integer := 0
                );
                port (
                    x : in signed(DATA_WIDTH-1 downto 0);
                    y : out signed(DATA_WIDTH-1 downto 0)
                );
            end component tanh;

        begin

            clock_process: process
            begin
                clk <= '0';
                wait for clk_period/2;
                clk <= '1';
                wait for clk_period/2;
            end process clock_process;

            uut: tanh
            port map (
                x => test_input,
                y => test_output
            );

            test_process: process
            begin
                report "======Simulation Start======" severity Note;

                test_input <= to_signed(-1281,16);
                wait for 1*clk_period;
                report "The value of 'test_output' is " & integer'image(to_integer(unsigned(test_output)));
                assert test_output=-256 report "The test case -1281 fail" severity failure;

                test_input <= to_signed(-1000,16);
                wait for 1*clk_period;
                report "The value of 'test_output' is " & integer'image(to_integer(unsigned(test_output)));
                assert test_output=-255 report "The test case -1000 fail" severity failure;

                test_input <= to_signed(-500,16);
                wait for 1*clk_period;
                report "The value of 'test_output' is " & integer'image(to_integer(unsigned(test_output)));
                assert test_output=-246 report "The test case -500 fail" severity failure;

                test_input <= to_signed(0,16);
                wait for 1*clk_period;
                report "The value of 'test_output' is " & integer'image(to_integer(unsigned(test_output)));
                assert test_output=0 report "The test case 0 fail" severity failure;

                test_input <= to_signed(500,16);
                wait for 1*clk_period;
                report "The value of 'test_output' is " & integer'image(to_integer(unsigned(test_output)));
                assert test_output=245 report "The test case 500 fail" severity failure;

                test_input <= to_signed(800,16);
                wait for 1*clk_period;
                report "The value of 'test_output' is " & integer'image(to_integer(unsigned(test_output)));
                assert test_output=254 report "The test case 800 fail" severity failure;

                test_input <= to_signed(1024,16);
                wait for 1*clk_period;
                report "The value of 'test_output' is " & integer'image(to_integer(unsigned(test_output)));
                assert test_output=255 report "The test case 1024 fail" severity failure;


                report "======Simulation Success======" severity Note;
                report "Please check the output message." severity Note;

                wait;

            end process test_process;

        end architecture rtl;
        """
        to_fp = partial(float_values_to_fixed_point, total_bits=16, frac_bits=0)
        tanh = PrecomputedScalarTestBench(
            component_name="tanh",
            x_list_for_testing=to_fp([-1281, -1000, -500, 0, 500, 800, 1024]),
            y_list_for_testing=to_fp([-256, -255, -246, 0, 245, 254, 255]),
        )
        tanh_code = tanh()
        tanh_code_str = ""
        for line in tanh_code:
            tanh_code_str += line + "\n"
        self.check_generated_code(expected_code, tanh_code_str)
